from pandagg.tree.mapping import Mapping
from pandagg.interactive.mapping import IMapping
from pandagg.node.mapping.field_datatypes import (
    Text,
    Keyword,
    Long,
    Integer,
    Short,
    Byte,
    Double,
    HalfFloat,
    ScaledFloat,
    Date,
    DateNanos,
    Boolean,
    Binary,
    IntegerRange,
    Float,
    FloatRange,
    LongRange,
    DoubleRange,
    DateRange,
    Object,
    Nested,
    GeoPoint,
    GeoShape,
    IP,
    Completion,
    TokenCount,
    MapperMurMur3,
    MapperAnnotatedText,
    Percolator,
    Join,
    RankFeature,
    RankFeatures,
    DenseVector,
    SparseVector,
    SearchAsYouType,
    Alias,
    Flattened,
    Shape,
    Histogram,
)
from pandagg.node.mapping.meta_fields import (
    Index,
    Type,
    Id,
    FieldNames,
    Source,
    Size,
    Ignored,
    Routing,
    Meta,
)

__all__ = [
    "Mapping",
    "IMapping",
    "Text",
    "Keyword",
    "Long",
    "Integer",
    "Short",
    "Byte",
    "Double",
    "HalfFloat",
    "ScaledFloat",
    "Date",
    "DateNanos",
    "Boolean",
    "Binary",
    "IntegerRange",
    "Float",
    "FloatRange",
    "LongRange",
    "DoubleRange",
    "DateRange",
    "Object",
    "Nested",
    "GeoPoint",
    "GeoShape",
    "IP",
    "Completion",
    "TokenCount",
    "MapperMurMur3",
    "MapperAnnotatedText",
    "Percolator",
    "Join",
    "RankFeature",
    "RankFeatures",
    "DenseVector",
    "SparseVector",
    "SearchAsYouType",
    "Alias",
    "Flattened",
    "Shape",
    "Histogram",
    "Index",
    "Type",
    "Id",
    "FieldNames",
    "Source",
    "Size",
    "Ignored",
    "Routing",
    "Meta",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import copy

from builtins import str as text

from future.utils import iterkeys, iteritems

from pandagg.interactive.response import IResponse
from pandagg.node.aggs.abstract import UniqueBucketAgg, MetricAgg, Root
from pandagg.node.aggs.bucket import Nested
from pandagg.tree.response import AggsResponseTree


class Response:
    def __init__(self, data, search):
        self.data = data
        self.__search = search

        self.took = data["took"]
        self.timed_out = data["timed_out"]
        self._shards = data["_shards"]
        self.hits = Hits(data["hits"])
        self.aggregations = Aggregations(
            data.get("aggregations", {}), search=self.__search
        )
        self.profile = data.get("profile")

    def __iter__(self):
        return iter(self.hits)

    @property
    def success(self):
        return (
            self._shards["total"] == self._shards["successful"] and not self.timed_out
        )

    def __len__(self):
        return len(self.hits)

    def __repr__(self):
        return (
            "<Response> took %dms, success: %s, total result %s, contains %s hits"
            % (self.took, self.success, self.hits._total_repr(), len(self.hits))
        )


class Hits:
    def __init__(self, hits):
        self.data = hits
        self.total = hits["total"]
        self.hits = [Hit(hit) for hit in hits.get("hits", [])]
        self.max_score = hits["max_score"]

    def __len__(self):
        return len(self.hits)

    def __iter__(self):
        return iter(self.hits)

    def _total_repr(self):
        if not isinstance(self.total, dict):
            return str(self.total)
        if self.total.get("relation") == "eq":
            return str(self.total["value"])
        if self.total.get("relation") == "gte":
            return ">=%d" % self.total["value"]
        raise ValueError("Invalid total %s" % self.total)

    def to_dataframe(self, expand_source=True, source_only=True):
        """
        Return hits as pandas dataframe.
        Requires pandas dependency.
        :param expand_source: if True, `_source` sub-fields are expanded as columns
        :param source_only: if True, doesn't include hit metadata (except id which is used as dataframe index)
        """
        try:
            import pandas as pd
        except ImportError:
            raise ImportError(
                'Using dataframe output format requires to install pandas. Please install "pandas" or '
                "use another output format."
            )
        hits = self.data.get("hits", [])
        if not hits:
            return pd.DataFrame()
        if not expand_source:
            return pd.DataFrame(hits).set_index("_id")
        flattened_hits = []
        for hit in hits:
            hit_metadata = hit.copy()
            hit_source = hit_metadata.pop("_source")
            if source_only:
                hit_source["_id"] = hit_metadata["_id"]
            else:
                hit_source.update(hit_metadata)
            flattened_hits.append(hit_source)
        return pd.DataFrame(flattened_hits).set_index("_id")

    def __repr__(self):
        if not isinstance(self.total, dict):
            total_repr = text(self.total)
        elif self.total.get("relation") == "eq":
            total_repr = text(self.total["value"])
        elif self.total.get("relation") == "gte":
            total_repr = ">%d" % self.total["value"]
        else:
            raise ValueError("Invalid total %s" % self.total)
        return "<Hits> total: %s, contains %d hits" % (total_repr, len(self.hits))


class Hit:
    def __init__(self, data):
        self.data = data
        self._source = data.get("_source")
        self._score = data.get("_score")
        self._id = data.get("_id")
        self._type = data.get("_type")
        self._index = data.get("_index")

    def __repr__(self):
        return "<Hit %s> score=%.2f" % (self._id, self._score)


class Aggregations:
    def __init__(self, data, search):
        self.data = data
        self.__search = search

    @property
    def _aggs(self):
        return self.__search._aggs

    @property
    def _query(self):
        return self.__search._query

    @property
    def _client(self):
        return self.__search._using

    @property
    def _index(self):
        return self.__search._index

    def keys(self):
        return self.data.keys()

    def get(self, key):
        return self.data[key]

    def _parse_group_by(
        self,
        response,
        until,
        row=None,
        agg_name=None,
        ancestors=None,
        row_as_tuple=False,
        with_single_bucket_groups=False,
    ):
        """Recursive parsing of succession of grouping aggregation clauses.

        Yields each row for which last bucket aggregation generated buckets.
        """
        # initialization: find ancestors once for faster computation
        if ancestors is None:
            until_id = self._aggs.id_from_key(until)
            ancestors = self._aggs.ancestors(until_id, include_current=True)
            # remove eventual fake root
            ancestors = [(k, n) for k, n in ancestors if k is not None]
            agg_name, agg_node = ancestors[-1]
        if not row:
            row = [] if row_as_tuple else {}
        if agg_name in response:
            agg_node = [n for k, n in ancestors if k == agg_name][0]
            for key, raw_bucket in agg_node.extract_buckets(response[agg_name]):
                sub_row = copy.copy(row)
                if (
                    not isinstance(agg_node, UniqueBucketAgg)
                    or with_single_bucket_groups
                ):
                    if row_as_tuple:
                        sub_row.append(key)
                    else:
                        sub_row[agg_name] = key
                if agg_name == until:
                    # end real yield
                    if row_as_tuple:
                        yield tuple(sub_row), raw_bucket
                    else:
                        yield sub_row, raw_bucket
                elif agg_name in {k for k, _ in ancestors}:
                    # yield children
                    for child_key, _ in self._aggs.children(agg_node.identifier):
                        for nrow, nraw_bucket in self._parse_group_by(
                            row=sub_row,
                            response=raw_bucket,
                            agg_name=child_key,
                            until=until,
                            row_as_tuple=row_as_tuple,
                            ancestors=ancestors,
                        ):
                            yield nrow, nraw_bucket

    def _normalize_buckets(self, agg_response, agg_name=None):
        """Recursive function to parse aggregation response as a normalized entities.
        Each response bucket is represented as a dict with keys (key, level, value, children)::

            {
                "level": "owner.id",
                "key": 35,
                "value": 235,
                "children": [
                ]
            }
        """
        agg_name = agg_name or self._aggs.root
        id_ = self._aggs.id_from_key(agg_name)
        agg_key, agg_node = self._aggs.get(id_)
        agg_children = self._aggs.children(agg_node.identifier)
        for key, raw_bucket in agg_node.extract_buckets(agg_response[agg_name]):
            result = {
                "level": agg_name,
                "key": key,
                "value": agg_node.extract_bucket_value(raw_bucket),
            }
            normalized_children = [
                normalized_child
                for child_key, child in agg_children
                for normalized_child in self._normalize_buckets(
                    agg_name=child_key, agg_response=raw_bucket
                )
            ]
            if normalized_children:
                result["children"] = normalized_children
            yield result

    def _grouping_agg(self, name=None):
        """Return aggregation node that used as grouping node.
        Note: in case there is only a nested aggregation below that node, group-by nested clause.
        """
        # if provided
        if name is not None:
            id_ = self._aggs.id_from_key(name)
            if not self._aggs._is_eligible_grouping_node(id_):
                raise ValueError(
                    "Cannot group by <%s>, not a valid grouping aggregation" % name
                )
            key, node = self._aggs.get(id_)
            # if parent of single nested clause and nested_autocorrect
            if self._aggs.nested_autocorrect:
                children = self._aggs.children(node.identifier)
                if len(children) == 1:
                    child_key, child_node = children[0]
                    if isinstance(child_node, Nested):
                        return child_key, child_node
            return key, node
        # if use of groupby method in Aggs class, use groupby pointer
        if self._aggs._groupby_ptr is not None:
            return self._aggs.get(self._aggs._groupby_ptr)

        id_ = self._aggs.deepest_linear_bucket_agg
        return self._aggs.get(id_)

    def to_tabular(
        self,
        index_orient=True,
        grouped_by=None,
        expand_columns=True,
        expand_sep="|",
        normalize=True,
        with_single_bucket_groups=False,
    ):
        """Build tabular view of ES response grouping levels (rows) until 'grouped_by' aggregation node included is
        reached, and using children aggregations of grouping level as values for each of generated groups (columns).

        Suppose an aggregation of this shape (A & B bucket aggregations)::

            A──> B──> C1
                 ├──> C2
                 └──> C3

        With grouped_by='B', breakdown ElasticSearch response (tree structure), into a tabular structure of this shape::

                                  C1     C2    C3
            A           B
            wood        blue      10     4     0
                        red       7      5     2
            steel       blue      1      9     0
                        red       23     4     2

        :param index_orient: if True, level-key samples are returned as tuples, else in a dictionnary
        :param grouped_by: name of the aggregation node used as last grouping level
        :param normalize: if True, normalize columns buckets
        :return: index, index_names, values
        """
        grouping_key, grouping_agg = self._grouping_agg(grouped_by)
        if grouping_key is None:
            index_values = [(tuple() if index_orient else dict(), self.data)]
            index_names = []
        else:
            index_names = [
                k
                for k, a in self._aggs.ancestors(
                    grouping_agg.identifier, from_root=True, include_current=True
                )
                if (not isinstance(a, UniqueBucketAgg) or with_single_bucket_groups)
                and k is not None
            ]
            index_values = list(
                self._parse_group_by(
                    response=self.data,
                    row_as_tuple=index_orient,
                    until=grouping_key,
                    with_single_bucket_groups=with_single_bucket_groups,
                )
            )
            if not index_values:
                return [], []

        if index_orient:
            rows = {
                row_index: self._serialize_columns(
                    row_values,
                    normalize=normalize,
                    total_agg=grouping_agg,
                    expand_columns=expand_columns,
                    expand_sep=expand_sep,
                )
                for row_index, row_values in index_values
            }
        else:
            rows = [
                dict(
                    row_index,
                    **self._serialize_columns(
                        row_values,
                        normalize=normalize,
                        total_agg=grouping_agg,
                        expand_columns=expand_columns,
                        expand_sep=expand_sep,
                    )
                )
                for row_index, row_values in index_values
            ]
        return index_names, rows

    def _serialize_columns(
        self, row_data, normalize, expand_columns, expand_sep, total_agg=None
    ):
        # extract value (usually 'doc_count') of grouping agg node
        result = {}
        if total_agg is not None and not isinstance(total_agg, Root):
            result[total_agg.VALUE_ATTRS[0]] = total_agg.extract_bucket_value(row_data)
            grouping_agg_children = self._aggs.children(total_agg.identifier)
        else:
            grouping_agg_children = self._aggs.children(self._aggs.root)

        # extract values of children, one columns per child
        for child_key, child in grouping_agg_children:
            if isinstance(child, (UniqueBucketAgg, MetricAgg)):
                result[child_key] = child.extract_bucket_value(row_data[child_key])
            elif expand_columns:
                for key, bucket in child.extract_buckets(row_data[child_key]):
                    result[
                        "%s%s%s" % (child_key, expand_sep, key)
                    ] = child.extract_bucket_value(bucket)
            elif normalize:
                result[child_key] = next(
                    self._normalize_buckets(row_data, child_key), None
                )
            else:
                result[child_key] = row_data[child_key]
        return result

    def to_dataframe(
        self, grouped_by=None, normalize_children=True, with_single_bucket_groups=False
    ):
        try:
            import pandas as pd
        except ImportError:
            raise ImportError(
                'Using dataframe output format requires to install pandas. Please install "pandas" or '
                "use another output format."
            )
        index_names, rows = self.to_tabular(
            index_orient=True,
            grouped_by=grouped_by,
            normalize=normalize_children,
            with_single_bucket_groups=with_single_bucket_groups,
        )
        index, values = zip(*iteritems(rows))
        if not index:
            return pd.DataFrame()
        if len(index[0]) == 0:
            index = (None,) * len(index)
        else:
            index = pd.MultiIndex.from_tuples(index, names=index_names)
        return pd.DataFrame(index=index, data=list(values))

    def to_normalized(self):
        children = []
        for k in sorted(iterkeys(self.data)):
            for child in self._normalize_buckets(self.data, k):
                children.append(child)
        return {"level": "root", "key": None, "value": None, "children": children}

    def to_tree(self):
        return AggsResponseTree(aggs=self._aggs).parse(self.data)

    def to_interactive_tree(self):
        return IResponse(tree=self.to_tree(), search=self.__search, depth=1)

    def serialize(self, output="tabular", **kwargs):
        """
        :param output: output format, one of "raw", "tree", "interactive_tree", "normalized", "tabular", "dataframe"
        :param kwargs: tabular serialization kwargs
        :return:
        """
        if output == "raw":
            return self.data
        elif output == "tree":
            return self.to_tree()
        elif output == "interactive_tree":
            return self.to_interactive_tree()
        elif output == "normalized":
            return self.to_normalized()
        elif output == "tabular":
            return self.to_tabular(**kwargs)
        elif output == "dataframe":
            return self.to_dataframe(**kwargs)
        else:
            raise NotImplementedError("Unkown %s output format." % output)

    def __repr__(self):
        if not self.keys():
            return "<Aggregations> empty"
        return "<Aggregations> %s" % list(map(text, self.keys()))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from collections import OrderedDict, defaultdict

from future.utils import iteritems

from pandagg.node.query.joining import Nested
from pandagg.tree._tree import Tree

from pandagg.node.response.bucket import Bucket, BucketNode
from pandagg.tree.query.abstract import Query


class AggsResponseTree(Tree):
    """Tree representation of an ElasticSearch response."""

    node_class = BucketNode

    def __init__(self, aggs, raw_response=None):
        """
        :param aggs: instance of pandagg.agg.Agg from which this Elasticsearch response originates.
        """
        super(AggsResponseTree, self).__init__()
        self.__aggs = aggs
        root_node = BucketNode()
        self.insert_node(root_node)
        if raw_response:
            self.parse(raw_response)

    def _clone_init(self, deep=False):
        return AggsResponseTree(aggs=self.__aggs.clone(deep=deep))

    def parse(self, raw_response):
        """Build response tree from ElasticSearch aggregation response

        :param raw_response: ElasticSearch aggregation response
        :return: self
        """
        _, agg_root_node = self.__aggs.get(self.__aggs.root)
        for child_name, child in self.__aggs.children(agg_root_node.identifier):
            self._parse_node_with_children(
                child_name, agg_node=child, raw_response=raw_response, pid=self.root
            )
        return self

    def _parse_node_with_children(self, agg_name, agg_node, raw_response, pid):
        """Recursive method to parse ES raw response.

        :param agg_node: current aggregation, pandagg.nodes.AggNode instance
        :param raw_response: ES response at current level, dict
        :param pid: parent node identifier
        """
        agg_raw_response = raw_response.get(agg_name)
        for key, raw_value in agg_node.extract_buckets(agg_raw_response):
            bucket = Bucket(
                level=agg_name, key=key, value=agg_node.extract_bucket_value(raw_value)
            )
            self.insert_node(bucket, parent_id=pid)
            for child_name, child in self.__aggs.children(agg_node.identifier):
                self._parse_node_with_children(
                    child_name,
                    agg_node=child,
                    raw_response=raw_value,
                    pid=bucket.identifier,
                )

    def bucket_properties(self, bucket, properties=None, end_level=None, depth=None):
        """Recursive method returning a given bucket's properties in the form of an ordered dictionnary.
        Travel from current bucket through all ancestors until reaching root.

        :param bucket: instance of pandagg.buckets.buckets.Bucket
        :param properties: OrderedDict accumulator of 'level' -> 'key'
        :param end_level: optional parameter to specify until which level properties are fetched
        :param depth: optional parameter to specify a limit number of levels which are fetched
        :return: OrderedDict of structure 'level' -> 'key'
        """
        if properties is None:
            properties = OrderedDict()
        if bucket.level is not None:
            properties[bucket.level] = bucket.key
        if depth is not None:
            depth -= 1
        _, parent = self.parent(bucket.identifier)
        if bucket.level == end_level or depth == 0 or parent is None:
            return properties
        return self.bucket_properties(parent, properties, end_level, depth)

    @classmethod
    def _build_filter(
        cls, nid_to_children, filters_per_nested_level, current_nested_path=None
    ):
        """Recursive function to build bucket filters from highest to deepest nested conditions."""
        current_conditions = filters_per_nested_level.get(current_nested_path, [])
        nested_children = nid_to_children[current_nested_path]
        for nested_child in nested_children:
            nested_child_conditions = cls._build_filter(
                nid_to_children=nid_to_children,
                filters_per_nested_level=filters_per_nested_level,
                current_nested_path=nested_child,
            )
            if nested_child_conditions:
                current_conditions.append(
                    Nested(path=nested_child, query=nested_child_conditions)
                )
        q = Query()
        for clause in current_conditions:
            q = q.query(clause)
        return q

    def get_bucket_filter(self, nid):
        """Build query filtering documents belonging to that bucket.
        Suppose the following configuration::

            Base                        <- filter on base
              |── Nested_A                 no filter on A (nested still must be applied for children)
              |     |── SubNested A1
              |     └── SubNested A2    <- filter on A2
              └── Nested_B              <- filter on B

        """
        tree_mapping = self.__aggs.mapping

        b_key, selected_bucket = self.get(nid)
        bucket_properties = self.bucket_properties(selected_bucket)
        agg_node_key_tuples = [
            (self.__aggs.get(self.__aggs.id_from_key(level))[1], key)
            for level, key in iteritems(bucket_properties)
        ]

        filters_per_nested_level = defaultdict(list)

        for agg_node, key in agg_node_key_tuples:
            level_agg_filter = agg_node.get_filter(key)
            # remove unnecessary match_all filters
            if level_agg_filter is not None and "match_all" not in level_agg_filter:
                current_nested = self.__aggs.applied_nested_path_at_node(
                    agg_node.identifier
                )
                filters_per_nested_level[current_nested].append(level_agg_filter)

        nested_with_conditions = [n for n in filters_per_nested_level.keys() if n]

        all_nesteds = [
            n.identifier
            for n in tree_mapping.list(
                filter_=lambda x: (x.KEY == "nested")
                and any((i in x.identifier or "" for i in nested_with_conditions))
            )
        ]

        nid_to_children = defaultdict(set)
        for nested in all_nesteds:
            nested_with_parents = [
                n
                for n in tree_mapping.ancestors(nid=nested, id_only=False)
                if n.KEY == "nested"
            ]
            nearest_nested_parent = next(iter(nested_with_parents[1:]), None)
            nid_to_children[nearest_nested_parent].add(nested)
        return self._build_filter(nid_to_children, filters_per_nested_level).to_dict()

    def show(self, **kwargs):
        kwargs["key"] = kwargs.get("key", lambda x: x.line_repr(depth=0))
        return super(AggsResponseTree, self).show(**kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Not implemented aggregations include:
- children agg
- geo-distance
- geo-hash grid
- ipv4
- sampler
- significant terms
"""

from builtins import str as text

from pandagg.node.types import NUMERIC_TYPES
from pandagg.node.aggs.abstract import MultipleBucketAgg, UniqueBucketAgg


class Global(UniqueBucketAgg):

    KEY = "global"
    VALUE_ATTRS = ["doc_count"]

    def __init__(self, meta=None):
        super(Global, self).__init__(agg_body={}, meta=meta)

    def get_filter(self, key):
        return None


class Filter(UniqueBucketAgg):

    KEY = "filter"
    VALUE_ATTRS = ["doc_count"]

    def __init__(self, filter=None, meta=None, **kwargs):
        if (filter is not None) != (not kwargs):
            raise ValueError(
                'Filter aggregation requires exactly one of "filter" or "kwargs"'
            )
        if filter:
            filter_ = filter.copy()
        else:
            filter_ = kwargs.copy()
        self.filter = filter_
        super(Filter, self).__init__(meta=meta, **filter_)

    def get_filter(self, key):
        return self.filter


class MatchAll(Filter):
    def __init__(self, meta=None):
        super(MatchAll, self).__init__(filter={"match_all": {}}, meta=meta)


class Nested(UniqueBucketAgg):

    KEY = "nested"
    VALUE_ATTRS = ["doc_count"]
    WHITELISTED_MAPPING_TYPES = ["nested"]

    def __init__(self, path, meta=None, **body):
        self.path = path
        super(Nested, self).__init__(path=path, meta=meta, **body)

    def get_filter(self, key):
        return None


class ReverseNested(UniqueBucketAgg):

    KEY = "reverse_nested"
    VALUE_ATTRS = ["doc_count"]
    WHITELISTED_MAPPING_TYPES = ["nested"]

    def __init__(self, path=None, meta=None, **body):
        self.path = path
        body_kwargs = dict(body)
        if path:
            body_kwargs["path"] = path
        super(ReverseNested, self).__init__(meta=meta, **body_kwargs)

    def get_filter(self, key):
        return None


class Missing(UniqueBucketAgg):
    KEY = "missing"
    VALUE_ATTRS = ["doc_count"]
    BLACKLISTED_MAPPING_TYPES = []

    def __init__(self, field, meta=None, **body):
        super(UniqueBucketAgg, self).__init__(field=field, meta=meta, **body)

    def get_filter(self, key):
        return {"bool": {"must_not": {"exists": {"field": self.field}}}}


class Terms(MultipleBucketAgg):
    """Terms aggregation."""

    KEY = "terms"
    VALUE_ATTRS = ["doc_count", "doc_count_error_upper_bound", "sum_other_doc_count"]
    BLACKLISTED_MAPPING_TYPES = []

    def __init__(self, field, missing=None, size=None, meta=None, **body):
        self.field = field
        self.missing = missing
        self.size = size

        body_kwargs = dict(body)
        if missing is not None:
            body_kwargs["missing"] = missing
        if size is not None:
            body_kwargs["size"] = size

        super(Terms, self).__init__(field=field, meta=meta, **body_kwargs)

    def get_filter(self, key):
        """Provide filter to get documents belonging to document of given key."""
        if key == "missing":
            return {"bool": {"must_not": {"exists": {"field": self.field}}}}
        return {"term": {self.field: {"value": key}}}


class Filters(MultipleBucketAgg):

    KEY = "filters"
    VALUE_ATTRS = ["doc_count"]
    DEFAULT_OTHER_KEY = "_other_"
    IMPLICIT_KEYED = True

    def __init__(
        self, filters, other_bucket=False, other_bucket_key=None, meta=None, **body
    ):
        self.filters = filters
        self.other_bucket = other_bucket
        self.other_bucket_key = other_bucket_key
        body_kwargs = dict(body)
        if other_bucket:
            body_kwargs["other_bucket"] = other_bucket
        if other_bucket_key:
            body_kwargs["other_bucket_key"] = other_bucket_key

        super(Filters, self).__init__(filters=filters, meta=meta, **body_kwargs)

    def get_filter(self, key):
        """Provide filter to get documents belonging to document of given key."""
        if key in self.filters.keys():
            return self.filters[key]
        if self.other_bucket:
            if key == self.DEFAULT_OTHER_KEY or key == self.other_bucket_key:
                return {
                    "bool": {
                        "must_not": {"bool": {"should": list(self.filters.values())}}
                    }
                }
        raise ValueError("Unkown <%s> key" % key)


class Histogram(MultipleBucketAgg):

    KEY = "histogram"
    VALUE_ATTRS = ["doc_count"]
    WHITELISTED_MAPPING_TYPES = NUMERIC_TYPES

    def __init__(self, field, interval, meta=None, **body):
        self.field = field
        self.interval = interval
        super(Histogram, self).__init__(
            field=field, interval=interval, meta=meta, **body
        )

    def get_filter(self, key):
        try:
            key = float(key)
        except (TypeError, ValueError):
            raise ValueError(
                "Filter key of an histogram aggregation must be numeric, git <%s> of type <%s>"
                % (key, type(key))
            )
        return {"range": {self.field: {"gte": key, "lt": key + self.interval}}}


class DateHistogram(MultipleBucketAgg):
    KEY = "date_histogram"
    VALUE_ATTRS = ["doc_count"]
    WHITELISTED_MAPPING_TYPES = ["date"]
    # interval is deprecated from 7.2 in favor of calendar_interval and fixed interval

    def __init__(
        self,
        field,
        interval=None,
        calendar_interval=None,
        fixed_interval=None,
        meta=None,
        keyed=False,
        key_as_string=True,
        **body
    ):
        """Date Histogram aggregation.
        Note: interval is deprecated from 7.2 in favor of calendar_interval and fixed interval
        :param keyed: defines returned buckets format: if True, as dict.
        :param key_as_string: if True extracted key of bucket will be the formatted date (applicable if keyed=False)
        """
        self.field = field
        if not (interval or fixed_interval or calendar_interval):
            raise ValueError(
                'One of "interval", "calendar_interval" or "fixed_interval" must be provided.'
            )
        if interval:
            body["interval"] = interval
        if calendar_interval:
            body["calendar_interval"] = calendar_interval
        if fixed_interval:
            body["fixed_interval"] = fixed_interval
        self.interval = interval or calendar_interval or fixed_interval
        super(DateHistogram, self).__init__(
            field=field,
            meta=meta,
            keyed=keyed,
            key_path="key_as_string" if key_as_string else "key",
            **body
        )

    def get_filter(self, key):
        # https://www.elastic.co/guide/en/elasticsearch/reference/current/common-options.html#date-math
        return {
            "range": {self.field: {"gte": key, "lt": "%s||+%s" % (key, self.interval)}}
        }


class Range(MultipleBucketAgg):
    KEY = "range"
    VALUE_ATTRS = ["doc_count"]
    WHITELISTED_MAPPING_TYPES = NUMERIC_TYPES
    KEY_SEP = "-"

    def __init__(self, field, ranges, keyed=False, meta=None, **body):
        self.field = field
        self.ranges = ranges
        body_kwargs = dict(body)
        if keyed:
            self.bucket_key_suffix = "_as_string"
        else:
            self.bucket_key_suffix = None
        super(Range, self).__init__(
            field=field, ranges=ranges, meta=meta, keyed=keyed, **body_kwargs
        )

    @property
    def from_key(self):
        if self.bucket_key_suffix:
            return "from%s" % self.bucket_key_suffix
        return "from"

    @property
    def to_key(self):
        if self.bucket_key_suffix:
            return "to%s" % self.bucket_key_suffix
        return "to"

    def _extract_bucket_key(self, bucket):
        if self.from_key in bucket:
            key = "%s%s" % (bucket[self.from_key], self.KEY_SEP)
        else:
            key = "*-"
        if self.to_key in bucket:
            key += text(bucket[self.to_key])
        else:
            key += "*"
        return key

    def get_filter(self, key):
        from_, to_ = key.split(self.KEY_SEP)
        inner = {}
        if from_ != "*":
            inner["gte"] = from_
        if to_ != "*":
            inner["lt"] = to_
        return {"range": {self.field: inner}}


class DateRange(Range):
    KEY = "date_range"
    VALUE_ATTRS = ["doc_count"]
    WHITELISTED_MAPPING_TYPES = ["date"]
    # cannot use range '-' separator since some keys contain it
    KEY_SEP = "::"

    def __init__(self, field, key_as_string=True, meta=None, **body):
        self.key_as_string = key_as_string
        super(DateRange, self).__init__(field=field, keyed=True, meta=meta, **body)


class Composite(MultipleBucketAgg):
    KEY = "composite"

    def get_filter(self, key):
        raise NotImplementedError()



from unittest import TestCase

from pandagg.utils import equal_queries, equal_search


class PandaggTestCase(TestCase):
    def assertUnorderedEqual(self, first, second, msg=None):
        self.assertIsInstance(first, dict, msg)
        self.assertIsInstance(second, dict, msg)
        # preserve regular formatting
        if not equal_queries(first, second):
            self.assertEqual(first, second, msg)

    def assertQueryEqual(self, first, second, msg=None):
        self.assertUnorderedEqual(first, second, msg)

    def assertSearchEqual(self, first, second, msg=None):
        self.assertIsInstance(first, dict, msg)
        self.assertIsInstance(second, dict, msg)
        # preserve regular formatting
        if not equal_search(first, second):
            self.assertEqual(first, second, msg)



from __future__ import unicode_literals

from unittest import TestCase

from pandagg.node.query import (
    Terms,
    Term,
    Fuzzy,
    Exists,
    Ids,
    Prefix,
    Range,
    Regexp,
    TermsSet,
    Wildcard,
)


class TermLevelQueriesTestCase(TestCase):
    def test_fuzzy_clause(self):
        body = {"user": {"value": "ki"}}
        expected = {"fuzzy": body}

        q1 = Fuzzy(field="user", value="ki")
        q2 = Fuzzy(user="ki")
        q3 = Fuzzy(user={"value": "ki"})
        for q in (q1, q2, q3):
            self.assertEqual(q.body, body)
            self.assertEqual(q.to_dict(), expected)
            self.assertEqual(
                q.line_repr(depth=None), ("fuzzy", 'field=user, value="ki"')
            )

    def test_exists_clause(self):
        body = {"field": "user"}
        expected = {"exists": body}

        q = Exists(field="user")
        self.assertEqual(q.body, body)
        self.assertEqual(q.to_dict(), expected)
        self.assertEqual(q.line_repr(depth=None), ("exists", "field=user"))

    def test_ids_clause(self):
        body = {"values": [1, 4, 100]}
        expected = {"ids": body}

        q = Ids(values=[1, 4, 100])
        self.assertEqual(q.body, body)
        self.assertEqual(q.to_dict(), expected)
        self.assertEqual(q.line_repr(depth=None), ("ids", "values=[1, 4, 100]"))

    def test_prefix_clause(self):
        body = {"user": {"value": "ki"}}
        expected = {"prefix": body}

        q = Prefix(field="user", value="ki")
        self.assertEqual(q.body, body)
        self.assertEqual(q.to_dict(), expected)
        self.assertEqual(q.line_repr(depth=None), ("prefix", 'field=user, value="ki"'))

    def test_range_clause(self):
        body = {"age": {"gte": 10, "lte": 20, "boost": 2}}
        expected = {"range": body}

        q1 = Range(field="age", gte=10, lte=20, boost=2)
        q2 = Range(age={"gte": 10, "lte": 20, "boost": 2})
        for q in (q1, q2):
            self.assertEqual(q.body, body)
            self.assertEqual(q.to_dict(), expected)
            self.assertEqual(
                q.line_repr(depth=None), ("range", "field=age, boost=2, gte=10, lte=20")
            )

    def test_regexp_clause(self):
        body = {
            "user": {
                "value": "k.*y",
                "flags": "ALL",
                "max_determinized_states": 10000,
                "rewrite": "constant_score",
            }
        }
        expected = {"regexp": body}
        repr_ = (
            "regexp",
            'field=user, flags="ALL", max_determinized_states=10000, '
            'rewrite="constant_score", value="k.*y"',
        )

        q1 = Regexp(
            field="user",
            value="k.*y",
            flags="ALL",
            max_determinized_states=10000,
            rewrite="constant_score",
        )
        q2 = Regexp(
            user={
                "value": "k.*y",
                "flags": "ALL",
                "max_determinized_states": 10000,
                "rewrite": "constant_score",
            }
        )
        for q in (q1, q2):
            self.assertEqual(q.body, body)
            self.assertEqual(q.to_dict(), expected)
            self.assertEqual(q.line_repr(depth=None), repr_)

    def test_term_clause(self):
        body = {"user": {"value": "Kimchy", "boost": 1}}
        expected = {"term": body}

        q1 = Term(field="user", value="Kimchy", boost=1)
        q2 = Term(user={"value": "Kimchy", "boost": 1})
        for q in (q1, q2):
            self.assertEqual(q.body, body)
            self.assertEqual(q.to_dict(), expected)
            self.assertEqual(
                q.line_repr(depth=None), ("term", 'field=user, boost=1, value="Kimchy"')
            )

        # other format
        q3 = Term(user="Kimchy")
        self.assertEqual(q3.body, {"user": {"value": "Kimchy"}})
        self.assertEqual(q3.to_dict(), {"term": {"user": {"value": "Kimchy"}}})
        self.assertEqual(
            q3.line_repr(depth=None), ("term", 'field=user, value="Kimchy"')
        )

    def test_terms_clause(self):
        # note: != syntax than term (...), the "boost" parameter is at same level that "user"
        body = {"user": ["kimchy", "elasticsearch"], "boost": 1}
        expected = {"terms": body}

        q = Terms(user=["kimchy", "elasticsearch"], boost=1)
        self.assertEqual(q.body, body)
        self.assertEqual(q.to_dict(), expected)
        self.assertEqual(
            q.line_repr(depth=None),
            ("terms", 'boost=1, user=["kimchy", "elasticsearch"]'),
        )

    def test_terms_set_clause(self):
        body = {
            "programming_languages": {
                "terms": ["c++", "java", "php"],
                "minimum_should_match_field": "required_matches",
            }
        }
        expected = {"terms_set": body}

        q1 = TermsSet(
            field="programming_languages",
            terms=["c++", "java", "php"],
            minimum_should_match_field="required_matches",
        )
        q2 = TermsSet(
            programming_languages={
                "terms": ["c++", "java", "php"],
                "minimum_should_match_field": "required_matches",
            }
        )
        for q in (q1, q2):
            self.assertEqual(q.body, body)
            self.assertEqual(q.to_dict(), expected)
            self.assertEqual(
                q.line_repr(depth=None),
                (
                    "terms_set",
                    'field=programming_languages, minimum_should_match_field="required_matches", '
                    'terms=["c++", "java", "php"]',
                ),
            )

    def test_wildcard_clause(self):
        body = {"user": {"value": "ki*y", "boost": 1.0, "rewrite": "constant_score"}}
        expected = {"wildcard": body}

        q1 = Wildcard(field="user", value="ki*y", boost=1.0, rewrite="constant_score")
        q2 = Wildcard(user={"value": "ki*y", "boost": 1.0, "rewrite": "constant_score"})
        for q in (q1, q2):
            self.assertEqual(q.body, body)
            self.assertEqual(q.to_dict(), expected)
            self.assertEqual(
                q.line_repr(depth=None),
                (
                    "wildcard",
                    'field=user, boost=1.0, rewrite="constant_score", value="ki*y"',
                ),
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json

from lighttree import TreeBasedObj

from pandagg.tree.mapping import _mapping
from pandagg.interactive._field_agg_factory import field_classes_per_name
from pandagg.utils import DSLMixin


class IMapping(DSLMixin, TreeBasedObj):
    """Interactive wrapper upon mapping tree, allowing field navigation and quick access to single clause aggregations
    computation.
    """

    _REPR_NAME = "Mapping"
    _NODE_PATH_ATTR = "name"

    def __init__(
        self,
        mapping,
        client=None,
        index=None,
        depth=1,
        root_path=None,
        initial_tree=None,
    ):
        if mapping is None:
            raise ValueError("mapping cannot be None")
        self._client = client
        self._index = index
        super(IMapping, self).__init__(
            tree=_mapping(mapping),
            root_path=root_path,
            depth=depth,
            initial_tree=initial_tree,
        )
        # if we reached a leave, add aggregation capabilities based on reached mapping type
        self._set_agg_property_if_required()

    def _clone(self, nid, root_path, depth):
        return IMapping(
            self._tree.subtree(nid)[1],
            client=self._client,
            root_path=root_path,
            depth=depth,
            initial_tree=self._initial_tree,
            index=self._index,
        )

    def _set_agg_property_if_required(self):
        if self._client is not None and not self._tree.children(self._tree.root):
            _, field_node = self._tree.get(self._tree.root)
            if field_node.KEY in field_classes_per_name:
                search_class = self._get_dsl_type("search")
                self.a = field_classes_per_name[field_node.KEY](
                    search=search_class(
                        using=self._client,
                        index=self._index,
                        mapping=self._initial_tree,
                        repr_auto_execute=True,
                        nested_autocorrect=True,
                    ),
                    field=self._root_path,
                )

    def __call__(self, *args, **kwargs):
        print(
            json.dumps(
                self._tree.to_dict(), indent=2, sort_keys=True, separators=(",", ": ")
            )
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from lighttree import TreeBasedObj

from pandagg.tree.aggs.aggs import Aggs


class IResponse(TreeBasedObj):

    """Interactive aggregation response."""

    _ATTR = "attr_name"
    _COERCE_ATTR = True

    def __init__(self, tree, search, depth, root_path=None, initial_tree=None):
        self.__search = search
        super(IResponse, self).__init__(
            tree=tree, root_path=root_path, depth=depth, initial_tree=initial_tree
        )

    @property
    def _client(self):
        return self.__search._using

    @property
    def _index(self):
        return self.__search._index

    def _clone(self, nid, root_path, depth):
        return IResponse(
            tree=self._tree.subtree(nid)[1],
            root_path=root_path,
            depth=depth,
            initial_tree=self._initial_tree,
            search=self.__search,
        )

    def get_bucket_filter(self):
        """Build filters to select documents belonging to that bucket, independently from initial search query clauses."""
        return self._initial_tree.get_bucket_filter(self._tree.root)

    def search(self):
        # add bucket filter to initial query clauses
        s = self.__search.query(self.get_bucket_filter())
        # remove no-more necessary aggregations
        s._aggs = Aggs()
        return s



"""https://www.elastic.co/guide/en/elasticsearch/reference/current/mapping-types.html"""

from .abstract import ComplexField, RegularField


# CORE DATATYPES
# string
class Text(RegularField):
    KEY = "text"


class Keyword(RegularField):
    KEY = "keyword"


# numeric
class Long(RegularField):
    KEY = "long"


class Integer(RegularField):
    KEY = "integer"


class Short(RegularField):
    KEY = "short"


class Byte(RegularField):
    KEY = "byte"


class Double(RegularField):
    KEY = "double"


class Float(RegularField):
    KEY = "float"


class HalfFloat(RegularField):
    KEY = "half_float"


class ScaledFloat(RegularField):
    KEY = "scaled_float"


# date
class Date(RegularField):
    KEY = "date"


class DateNanos(RegularField):
    KEY = "date_nanos"


# boolean
class Boolean(RegularField):
    KEY = "boolean"


# binary
class Binary(RegularField):
    KEY = "binary"


# range
class IntegerRange(RegularField):
    KEY = "integer_range"


class FloatRange(RegularField):
    KEY = "float_range"


class LongRange(RegularField):
    KEY = "long_range"


class DoubleRange(RegularField):
    KEY = "double_range"


class DateRange(RegularField):
    KEY = "date_range"


# COMPLEX DATATYPES
class Object(ComplexField):
    KEY = "object"


class Nested(ComplexField):
    KEY = "nested"


# GEO DATATYPES
class GeoPoint(RegularField):
    """For lat/lon points"""

    KEY = "geo_point"


class GeoShape(RegularField):
    """For complex shapes like polygons"""

    KEY = "geo_shape"


# SPECIALIZED DATATYPES
class IP(RegularField):
    """for IPv4 and IPv6 addresses"""

    KEY = "IP"


class Completion(RegularField):
    """To provide auto-complete suggestions"""

    KEY = "completion"


class TokenCount(RegularField):
    """To count the number of tokens in a string"""

    KEY = "token_count"


class MapperMurMur3(RegularField):
    """To compute hashes of values at index-time and store them in the index"""

    KEY = "murmur3"


class MapperAnnotatedText(RegularField):
    """To index text containing special markup (typically used for identifying named entities)"""

    KEY = "annotated-text"


class Percolator(RegularField):
    """Accepts queries from the query-dsl"""

    KEY = "percolator"


class Join(RegularField):
    """Defines parent/child relation for documents within the same index"""

    KEY = "join"


class RankFeature(RegularField):
    """Record numeric feature to boost hits at query time."""

    KEY = "rank_feature"


class RankFeatures(RegularField):
    """Record numeric features to boost hits at query time."""

    KEY = "rank_features"


class DenseVector(RegularField):
    """Record dense vectors of float values."""

    KEY = "dense_vector"


class SparseVector(RegularField):
    """Record sparse vectors of float values."""

    KEY = "sparse_vector"


class SearchAsYouType(RegularField):
    """A text-like field optimized for queries to implement as-you-type completion"""

    KEY = "search_as_you_type"


class Alias(RegularField):
    """Defines an alias to an existing field."""

    KEY = "alias"


class Flattened(RegularField):
    """Allows an entire JSON object to be indexed as a single field."""

    KEY = "flattened"


class Shape(RegularField):
    """For arbitrary cartesian geometries."""

    KEY = "shape"


class Histogram(RegularField):
    """For pre-aggregated numerical values for percentiles aggregations."""

    KEY = "histogram"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from builtins import str as text

import json

from pandagg.node._node import Node


class AggNode(Node):
    """Wrapper around elasticsearch aggregation concept.
    https://www.elastic.co/guide/en/elasticsearch/reference/2.3/search-aggregations.html

    Each aggregation can be seen both a Node that can be encapsulated in a parent agg.

    Define a method to build aggregation request.
    """

    _type_name = "agg"
    KEY = None
    VALUE_ATTRS = None
    WHITELISTED_MAPPING_TYPES = None
    BLACKLISTED_MAPPING_TYPES = None

    def __init__(self, meta=None, **body):
        identifier = body.pop("identifier", None)
        self.body = body
        self.meta = meta
        self._children = {}
        super(AggNode, self).__init__(identifier=identifier)

    def line_repr(self, depth, **kwargs):
        # root node
        if self.KEY is None:
            return "_", ""
        repr_args = [text(self.KEY)]
        if self.body:
            repr_args.append(self._params_repr(self.body))
        unnamed = "<%s>" % ", ".join(repr_args)
        return "", unnamed

    @staticmethod
    def _params_repr(params):
        params = params or {}
        return ", ".join(
            "%s=%s" % (text(k), text(json.dumps(params[k], sort_keys=True)))
            for k in sorted(params.keys())
        )

    @classmethod
    def valid_on_field_type(cls, field_type):
        if cls.WHITELISTED_MAPPING_TYPES is not None:
            return field_type in cls.WHITELISTED_MAPPING_TYPES
        if cls.BLACKLISTED_MAPPING_TYPES is not None:
            return field_type not in cls.BLACKLISTED_MAPPING_TYPES
        return False

    def to_dict(self):
        """ElasticSearch aggregation queries follow this formatting::

            {
                "<aggregation_name>" : {
                    "<aggregation_type>" : {
                        <aggregation_body>
                    }
                    [,"meta" : {  [<meta_data_body>] } ]?
                }
            }

        to_dict() returns the following part (without aggregation name)::

            {
                "<aggregation_type>" : {
                    <aggregation_body>
                }
                [,"meta" : {  [<meta_data_body>] } ]?
            }
        """
        aggs = {self.KEY: self.body}
        if self.meta:
            aggs["meta"] = self.meta
        return aggs

    def get_filter(self, key):
        """Return filter query to list documents having this aggregation key.
        :param key: string
        :return: elasticsearch filter query
        """
        raise NotImplementedError()

    def extract_buckets(self, response_value):
        raise NotImplementedError()

    @classmethod
    def extract_bucket_value(cls, response, value_as_dict=False):
        attrs = cls.VALUE_ATTRS
        if value_as_dict:
            return {attr_: response.get(attr_) for attr_ in attrs}
        return response.get(attrs[0])

    def __str__(self):
        return "<{class_}, type={type}, body={body}>".format(
            class_=text(self.__class__.__name__),
            type=text(self.KEY),
            body=json.dumps(self.body),
        )

    def __eq__(self, other):
        if isinstance(other, AggNode):
            return other.to_dict() == self.to_dict()
        # make sure we still equal to a dict with the same data
        return other == self.to_dict()


class Root(AggNode):
    """Not a real aggregation. Just the initial empty dict."""

    KEY = "_root"

    def line_repr(self, depth, **kwargs):
        return "_", ""

    def extract_buckets(self, response_value):
        yield None, response_value

    @classmethod
    def extract_bucket_value(cls, response, value_as_dict=False):
        return None


class MetricAgg(AggNode):
    """Metric aggregation are aggregations providing a single bucket, with value attributes to be extracted."""

    VALUE_ATTRS = None

    def extract_buckets(self, response_value):
        yield None, response_value

    def get_filter(self, key):
        return None


class BucketAggNode(AggNode):
    """Bucket aggregation have special abilities: they can encapsulate other aggregations as children.
    Each time, the extracted value is a 'doc_count'.

    Provide methods:
    - to build aggregation request (with children aggregations)
    - to to extract buckets from raw response
    - to build query to filter documents belonging to that bucket

    Note: the aggs attribute's only purpose is for children initiation with the following syntax:
    >>> from pandagg.aggs import Terms, Avg
    >>> agg = Terms(
    >>>     field='some_path',
    >>>     aggs={
    >>>         'avg_agg': Avg(field='some_other_path')
    >>>     }
    >>> )
    """

    VALUE_ATTRS = None

    def __init__(self, meta=None, **body):
        identifier = body.pop("identifier", None)
        self.body = body
        self.meta = meta
        self._children = body.pop("aggs", None) or body.pop("aggregations", None) or {}
        super(AggNode, self).__init__(identifier=identifier)

    def extract_buckets(self, response_value):
        raise NotImplementedError()

    def get_filter(self, key):
        """Provide filter to get documents belonging to document of given key."""
        raise NotImplementedError()


class UniqueBucketAgg(BucketAggNode):
    """Aggregations providing a single bucket."""

    VALUE_ATTRS = None

    def extract_buckets(self, response_value):
        yield None, response_value

    def get_filter(self, key):
        raise NotImplementedError()


class MultipleBucketAgg(BucketAggNode):

    VALUE_ATTRS = None
    IMPLICIT_KEYED = False

    def __init__(self, keyed=None, key_path="key", meta=None, **body):
        """Aggregation that return either a list or a map of buckets.

        If keyed, ES buckets are expected as dict, else as list (in this case key_path is used to extract key from each
        list item).
        :param keyed:
        :param meta:
        :param aggs:
        :param body:
        """
        # keyed has another meaning in lighttree Node
        self.keyed_ = keyed or self.IMPLICIT_KEYED
        self.key_path = key_path
        if keyed and not self.IMPLICIT_KEYED:
            body["keyed"] = keyed
        super(MultipleBucketAgg, self).__init__(meta=meta, **body)

    def extract_buckets(self, response_value):
        buckets = response_value["buckets"]
        if self.keyed_:
            for key in sorted(buckets.keys()):
                yield (key, buckets[key])
        else:
            for bucket in buckets:
                yield (self._extract_bucket_key(bucket), bucket)

    def _extract_bucket_key(self, bucket):
        return bucket[self.key_path]

    def get_filter(self, key):
        raise NotImplementedError()


class FieldOrScriptMetricAgg(MetricAgg):
    """Metric aggregation based on single field."""

    VALUE_ATTRS = None

    def __init__(self, meta=None, **body):
        self.field = body.get("field")
        self.script = body.get("script")
        super(FieldOrScriptMetricAgg, self).__init__(meta=meta, **body)


class Pipeline(UniqueBucketAgg):

    VALUE_ATTRS = None

    def __init__(self, buckets_path, gap_policy=None, meta=None, **body):
        self.buckets_path = buckets_path
        self.gap_policy = gap_policy
        body_kwargs = dict(body)
        if gap_policy is not None:
            assert gap_policy in ("skip", "insert_zeros")
            body_kwargs["gap_policy"] = gap_policy

        super(Pipeline, self).__init__(meta=meta, buckets_path=buckets_path, **body)

    def get_filter(self, key):
        return None


class ScriptPipeline(Pipeline):
    KEY = None
    VALUE_ATTRS = "value"

    def __init__(self, script, buckets_path, gap_policy=None, meta=None, **body):
        super(ScriptPipeline, self).__init__(
            buckets_path=buckets_path,
            gap_policy=gap_policy,
            meta=meta,
            script=script,
            **body
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
from builtins import str as text


from pandagg.node._node import Node


class Field(Node):
    _type_name = "field"
    KEY = None

    def __init__(self, **body):
        super(Node, self).__init__()
        self._subfield = body.pop("_subfield", False)
        self._body = body.copy()

    def line_repr(self, depth, **kwargs):
        if self.KEY is None:
            return "_", ""
        return "", self._display_pattern % self.KEY.capitalize()

    @property
    def body(self):
        b = self._body.copy()
        if self.KEY in ("object", None):
            return b
        b["type"] = self.KEY
        return b

    @property
    def _display_pattern(self):
        if self.KEY == "object":
            return " {%s}"
        if self.KEY == "nested":
            return " [%s]"
        if self._subfield:
            return "~ %s"
        return "  %s"

    def __str__(self):
        return "<%s field>:\n%s" % (
            text(self.KEY).capitalize(),
            text(json.dumps(self.body, indent=4)),
        )


class ComplexField(Field):
    KEY = None

    def __init__(self, **body):
        properties = body.pop("properties", None)
        if properties and not isinstance(properties, dict):
            raise ValueError("Invalid properties %s" % properties)
        self.properties = properties or {}
        super(ComplexField, self).__init__(**body)


class RegularField(Field):
    KEY = None

    def __init__(self, **body):
        fields = body.pop("fields", None)
        if fields and not isinstance(fields, dict):
            raise ValueError("Invalid fields %s" % fields)
        self.fields = fields
        super(RegularField, self).__init__(**body)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from future.utils import iteritems

from pandagg.node.mapping.abstract import Field, RegularField, ComplexField


from pandagg.exceptions import (
    AbsentMappingFieldError,
    InvalidOperationMappingFieldError,
)
from pandagg.tree._tree import Tree


def _mapping(m):
    if m is None:
        return None
    if isinstance(m, dict):
        return Mapping(**m)
    if isinstance(m, Mapping):
        return m
    raise TypeError("Unsupported %s type for Mapping" % type(m))


class Mapping(Tree):

    node_class = Field
    KEY = None

    def __init__(self, properties=None, dynamic=False, **kwargs):
        """"""
        super(Mapping, self).__init__()
        root_node = Field(dynamic=dynamic, **kwargs)
        self.insert_node(root_node)
        if properties:
            self._insert(root_node.identifier, properties, False)

    def _insert(self, pid, el, is_subfield):
        if not isinstance(el, dict):
            raise ValueError("Wrong declaration, got %s" % el)
        for name, field in iteritems(el):
            if isinstance(field, dict):
                field = field.copy()
                field = Field._get_dsl_class(field.pop("type", "object"))(
                    _subfield=is_subfield, **field
                )
            elif isinstance(field, Field):
                field._subfield = is_subfield
                pass
            else:
                raise ValueError("Unsupported type %s" % type(field))
            self.insert_node(field, key=name, parent_id=pid)
            if isinstance(field, ComplexField) and field.properties:
                self._insert(field.identifier, field.properties, False)
            if isinstance(field, RegularField) and field.fields:
                if is_subfield:
                    raise ValueError(
                        "Cannot insert subfields into a subfield on field %s" % name
                    )
                self._insert(field.identifier, field.fields, True)

    def to_dict(self, from_=None, depth=None):
        if self.root is None:
            return None
        from_ = self.root if from_ is None else from_
        key, node = self.get(from_)
        children_queries = {}
        if depth is None or depth > 0:
            if depth is not None:
                depth -= 1
            for child_key, child_node in self.children(node.identifier):
                children_queries[child_key] = self.to_dict(
                    from_=child_node.identifier, depth=depth
                )
        serialized_node = node.body
        if children_queries:
            if node.KEY is None or node.KEY in ("object", "nested"):
                serialized_node["properties"] = children_queries
            else:
                serialized_node["fields"] = children_queries
        return serialized_node

    def validate_agg_node(self, agg_node, exc=True):
        """Ensure if node has field or path that it exists in mapping, and that required aggregation type
        if allowed on this kind of field.
        :param agg_node: AggNode you want to validate on this mapping
        :param exc: boolean, if set to True raise exception if invalid
        :rtype: boolean
        """
        if hasattr(agg_node, "path"):
            if agg_node.path is None:
                # reverse nested
                return True
            return self.resolve_path_to_id(agg_node.path) in self

        if not hasattr(agg_node, "field"):
            return True

        # TODO take into account flattened data type
        try:
            nid = self.get_node_id_by_path(agg_node.field)
        except StopIteration:
            raise AbsentMappingFieldError(
                u"Agg of type <%s> on non-existing field <%s>."
                % (agg_node.KEY, agg_node.field)
            )
        _, field = self.get(nid)

        field_type = field.KEY
        if not agg_node.valid_on_field_type(field_type):
            if not exc:
                return False
            raise InvalidOperationMappingFieldError(
                u"Agg of type <%s> not possible on field of type <%s>."
                % (agg_node.KEY, field_type)
            )
        return True

    def mapping_type_of_field(self, field_path):
        try:
            _, node = self.get(field_path, by_path=True)
            return node.KEY
        except Exception:
            raise AbsentMappingFieldError(
                u"<%s field is not present in mapping>" % field_path
            )

    def nested_at_field(self, field_path):
        nesteds = self.list_nesteds_at_field(field_path)
        if nesteds:
            return nesteds[0]
        return None

    def list_nesteds_at_field(self, field_path):
        """List nested paths that apply at a given path."""
        path_nid = self.get_node_id_by_path(field_path)
        # from deepest to highest
        return [
            self.get_path(nid)
            for nid in self.ancestors_ids(path_nid, include_current=True)
            if self.get(nid)[1].KEY == "nested"
        ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging

from logging import NullHandler

# ensures all required classes are registered in DSLMeta
from .interactive import *
from .search import *

# Inspired by https://python-guide-pt-br.readthedocs.io/fr/latest/writing/logging.html#logging-in-a-library
# Set default logging handler to avoid "No handler found" warnings.
logging.getLogger(__name__).addHandler(NullHandler())

__all__ = []



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Pipeline aggregations:
https://www.elastic.co/guide/en/elasticsearch/reference/2.3/search-aggregations-pipeline.html
"""

from pandagg.node.aggs.abstract import Pipeline, ScriptPipeline


class AvgBucket(Pipeline):
    KEY = "avg_bucket"
    VALUE_ATTRS = ["value"]


class Derivative(Pipeline):
    KEY = "derivative"
    VALUE_ATTRS = ["value"]


class MaxBucket(Pipeline):
    KEY = "max_bucket"
    VALUE_ATTRS = ["value"]


class MinBucket(Pipeline):
    KEY = "min_bucket"
    VALUE_ATTRS = ["value"]


class SumBucket(Pipeline):
    KEY = "sum_bucket"
    VALUE_ATTRS = ["value"]


class StatsBucket(Pipeline):
    KEY = "stats_bucket"
    VALUE_ATTRS = ["count", "min", "max", "avg", "sum"]


class ExtendedStatsBucket(Pipeline):
    KEY = "extended_stats_bucket"
    VALUE_ATTRS = [
        "count",
        "min",
        "max",
        "avg",
        "sum",
        "sum_of_squares",
        "variance",
        "std_deviation",
        "std_deviation_bounds",
    ]


class PercentilesBucket(Pipeline):
    KEY = "percentiles_bucket"
    VALUE_ATTRS = ["values"]


class MovingAvg(Pipeline):
    KEY = "moving_avg"
    VALUE_ATTRS = ["value"]


class CumulativeSum(Pipeline):
    KEY = "cumulative_sum"
    VALUE_ATTRS = ["value"]


class BucketScript(ScriptPipeline):
    KEY = "bucket_script"
    VALUE_ATTRS = ["value"]


class BucketSelector(ScriptPipeline):
    KEY = "bucket_selector"
    VALUE_ATTRS = None


class BucketSort(ScriptPipeline):
    KEY = "bucket_sort"
    VALUE_ATTRS = None


class SerialDiff(Pipeline):
    KEY = "serial_diff"
    VALUE_ATTRS = ["value"]



from pandagg.tree.query.abstract import Query
from pandagg.node.query.shape import Shape
from pandagg.node.query.term_level import (
    Exists,
    Fuzzy,
    Ids,
    Prefix,
    Range,
    Regexp,
    Term,
    Terms,
    TermsSet,
    Type,
    Wildcard,
)
from pandagg.node.query.full_text import (
    Intervals,
    Match,
    MatchBoolPrefix,
    MatchPhrase,
    MatchPhrasePrefix,
    MultiMatch,
    Common,
    QueryString,
    SimpleQueryString,
)
from pandagg.node.query.compound import (
    Bool,
    Boosting,
    ConstantScore,
    FunctionScore,
    DisMax,
)
from pandagg.node.query.joining import Nested, HasChild, HasParent, ParentId
from pandagg.node.query.geo import GeoShape, GeoPolygone, GeoDistance, GeoBoundingBox
from pandagg.node.query.specialized import (
    DistanceFeature,
    MoreLikeThis,
    Percolate,
    RankFeature,
    Script,
    Wrapper,
)
from pandagg.node.query.specialized_compound import ScriptScore, PinnedQuery

__all__ = [
    "Query",
    # term level
    "Exists",
    "Fuzzy",
    "Ids",
    "Prefix",
    "Range",
    "Regexp",
    "Term",
    "Terms",
    "TermsSet",
    "Type",
    "Wildcard",
    # full text
    "Intervals",
    "Match",
    "MatchBoolPrefix",
    "MatchPhrase",
    "MatchPhrasePrefix",
    "MultiMatch",
    "Common",
    "QueryString",
    "SimpleQueryString",
    # compound
    "Bool",
    "Boosting",
    "ConstantScore",
    "FunctionScore",
    "DisMax",
    # joining
    "Nested",
    "HasParent",
    "HasChild",
    "ParentId",
    # shape
    "Shape",
    # geo
    "GeoShape",
    "GeoPolygone",
    "GeoDistance",
    "GeoBoundingBox",
    # specialized
    "DistanceFeature",
    "MoreLikeThis",
    "Percolate",
    "RankFeature",
    "Script",
    "Wrapper",
    "ScriptScore",
    "PinnedQuery",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from tests import PandaggTestCase
from lighttree.exceptions import NotFoundNodeError
import pandas as pd

from pandagg.search import Search
from pandagg.tree.response import AggsResponseTree
from pandagg.response import Response, Hits, Hit, Aggregations
from pandagg.tree.aggs.aggs import Aggs

import tests.testing_samples.data_sample as sample
from pandagg.utils import ordered
from tests.testing_samples.mapping_example import MAPPING


class ResponseTestCase(PandaggTestCase):
    def test_hit(self):
        h = Hit(
            {
                "_index": "my_index_01",
                "_type": "_doc",
                "_id": "1",
                "_score": 1.0,
                "_source": {"field_23": 1},
            }
        )
        self.assertEqual(h._index, "my_index_01")
        self.assertEqual(h._type, "_doc")
        self.assertEqual(h._id, "1")
        self.assertEqual(h._score, 1.0)
        self.assertEqual(h._source, {"field_23": 1})

        self.assertEqual(h.__repr__(), "<Hit 1> score=1.00")

    def test_hits(self):
        hits = Hits(
            {
                "total": {"value": 34, "relation": "eq"},
                "max_score": 0.0,
                "hits": [
                    {
                        "_index": "my_index_01",
                        "_type": "_doc",
                        "_id": "1",
                        "_score": 1.0,
                        "_source": {"field_23": 1, "yo": 2},
                    },
                    {
                        "_index": "my_index_01",
                        "_type": "_doc",
                        "_id": "2",
                        "_score": 0.2,
                        "_source": {"field_23": 2, "to": 1},
                    },
                ],
            }
        )
        self.assertEqual(hits.total, {"value": 34, "relation": "eq"})
        self.assertEqual(hits._total_repr(), "34")
        self.assertEqual(len(hits), 2)
        self.assertEqual(len(hits.hits), 2)
        for h in hits.hits:
            self.assertIsInstance(h, Hit)
        self.assertEqual(hits.__repr__(), "<Hits> total: 34, contains 2 hits")

        # as dataframe
        source_only_df = hits.to_dataframe()
        with_metadata_df = hits.to_dataframe(source_only=False)
        non_expanded_df = hits.to_dataframe(expand_source=False)

        self.assertEqual(
            sorted(source_only_df.columns.tolist()), sorted(["field_23", "to", "yo"])
        )
        self.assertEqual(source_only_df.shape, (2, 3))
        self.assertEqual(source_only_df.index.tolist(), ["1", "2"])

        self.assertEqual(
            sorted(with_metadata_df.columns.tolist()),
            sorted(["_index", "_score", "_type", "field_23", "to", "yo"]),
        )
        self.assertEqual(with_metadata_df.shape, (2, 6))
        self.assertEqual(with_metadata_df.index.tolist(), ["1", "2"])

        self.assertEqual(
            sorted(non_expanded_df.columns.tolist()),
            ["_index", "_score", "_source", "_type"],
        )
        self.assertEqual(non_expanded_df.shape, (2, 4))
        self.assertEqual(non_expanded_df.index.tolist(), ["1", "2"])

    def test_response(self):
        r = Response(
            {
                "took": 42,
                "timed_out": False,
                "_shards": {"total": 10, "successful": 10, "skipped": 0, "failed": 0},
                "hits": {
                    "total": {"value": 34, "relation": "eq"},
                    "max_score": 0.0,
                    "hits": [
                        {
                            "_index": "my_index_01",
                            "_type": "_doc",
                            "_id": "1",
                            "_score": 1.0,
                            "_source": {"field_23": 1},
                        },
                        {
                            "_index": "my_index_01",
                            "_type": "_doc",
                            "_id": "2",
                            "_score": 0.2,
                            "_source": {"field_23": 2},
                        },
                    ],
                },
            },
            Search(),
        )
        self.assertEqual(r.took, 42)
        self.assertEqual(r.timed_out, False)
        self.assertEqual(r.success, True)
        self.assertEqual(
            r._shards, {"total": 10, "successful": 10, "skipped": 0, "failed": 0}
        )
        self.assertIsInstance(r.hits, Hits)
        self.assertEqual(
            r.__repr__(),
            "<Response> took 42ms, success: True, total result 34, contains 2 hits",
        )

        hits = list(iter(r))
        self.assertEqual(len(hits), 2)
        for hit in hits:
            self.assertIsInstance(hit, Hit)


class AggregationsResponseTestCase(PandaggTestCase):
    def test_parse_as_tree(self, *_):
        my_agg = Aggs(sample.EXPECTED_AGG_QUERY, mapping=MAPPING)
        response = Aggregations(
            data=sample.ES_AGG_RESPONSE, search=Search().aggs(my_agg)
        ).to_tree()
        self.assertIsInstance(response, AggsResponseTree)
        self.assertEqual(response.__str__(), sample.EXPECTED_RESPONSE_TREE_REPR)

    def test_normalize_buckets(self):
        my_agg = Aggs(sample.EXPECTED_AGG_QUERY, mapping=MAPPING)
        response = Aggregations(
            data=sample.ES_AGG_RESPONSE, search=Search().aggs(my_agg)
        ).to_normalized()
        self.assertEqual(
            ordered(response), ordered(sample.EXPECTED_NORMALIZED_RESPONSE)
        )

    def test_parse_as_tabular(self):
        # with single agg at root
        my_agg = Aggs(sample.EXPECTED_AGG_QUERY, mapping=MAPPING)
        index_names, index_values = Aggregations(
            data=sample.ES_AGG_RESPONSE, search=Search().aggs(my_agg)
        ).to_tabular(index_orient=True, grouped_by="global_metrics.field.name")

        self.assertEqual(
            index_names, ["classification_type", "global_metrics.field.name"]
        )
        self.assertEqual(
            index_values,
            {
                ("multilabel", "ispracticecompatible"): {
                    "avg_f1_micro": 0.72,
                    "avg_nb_classes": 18.71,
                    "doc_count": 128,
                },
                ("multilabel", "preservationmethods"): {
                    "avg_f1_micro": 0.8,
                    "avg_nb_classes": 9.97,
                    "doc_count": 76,
                },
                ("multiclass", "kind"): {
                    "avg_f1_micro": 0.89,
                    "avg_nb_classes": 206.5,
                    "doc_count": 370,
                },
                ("multiclass", "gpc"): {
                    "avg_f1_micro": 0.93,
                    "avg_nb_classes": 211.12,
                    "doc_count": 198,
                },
            },
        )

        # index_orient = False
        index_names, index_values = Aggregations(
            data=sample.ES_AGG_RESPONSE, search=Search().aggs(my_agg)
        ).to_tabular(index_orient=False, grouped_by="global_metrics.field.name")

        self.assertEqual(
            index_names, ["classification_type", "global_metrics.field.name"]
        )
        self.assertEqual(
            index_values,
            [
                {
                    "avg_f1_micro": 0.72,
                    "avg_nb_classes": 18.71,
                    "classification_type": "multilabel",
                    "doc_count": 128,
                    "global_metrics.field.name": "ispracticecompatible",
                },
                {
                    "avg_f1_micro": 0.8,
                    "avg_nb_classes": 9.97,
                    "classification_type": "multilabel",
                    "doc_count": 76,
                    "global_metrics.field.name": "preservationmethods",
                },
                {
                    "avg_f1_micro": 0.89,
                    "avg_nb_classes": 206.5,
                    "classification_type": "multiclass",
                    "doc_count": 370,
                    "global_metrics.field.name": "kind",
                },
                {
                    "avg_f1_micro": 0.93,
                    "avg_nb_classes": 211.12,
                    "classification_type": "multiclass",
                    "doc_count": 198,
                    "global_metrics.field.name": "gpc",
                },
            ],
        )

    def test_parse_as_tabular_multiple_roots(self):
        # with multiple aggs at root
        my_agg = Aggs(
            {
                "classification_type": {"terms": {"field": "classification_type"}},
                "avg_f1_score": {
                    "avg": {"field": "global_metrics.performance.test.micro.f1_score"}
                },
            }
        )

        raw_response = {
            "classification_type": {
                "doc_count_error_upper_bound": 0,
                "sum_other_doc_count": 0,
                "buckets": [
                    {"key": "multiclass", "doc_count": 439},
                    {"key": "multilabel", "doc_count": 433},
                ],
            },
            "avg_f1_score": {"value": 0.815},
        }
        index_names, index_values = Aggregations(
            data=raw_response, search=Search().aggs(my_agg)
        ).to_tabular(index_orient=True, expand_sep=" || ")

        self.assertEqual(index_names, [])
        self.assertEqual(
            index_values,
            {
                (): {
                    "avg_f1_score": 0.815,
                    "classification_type || multiclass": 439,
                    "classification_type || multilabel": 433,
                }
            },
        )

        # with specified grouped_by
        index_names, index_values = Aggregations(
            data=raw_response, search=Search().aggs(my_agg)
        ).to_tabular(grouped_by="classification_type")
        self.assertEqual(index_names, ["classification_type"])
        self.assertEqual(
            index_values,
            {("multiclass",): {"doc_count": 439}, ("multilabel",): {"doc_count": 433}},
        )

    def test_parse_as_dataframe(self):
        my_agg = Aggs(sample.EXPECTED_AGG_QUERY, mapping=MAPPING)
        df = Aggregations(
            data=sample.ES_AGG_RESPONSE, search=Search().aggs(my_agg)
        ).to_dataframe(grouped_by="global_metrics.field.name")
        self.assertIsInstance(df, pd.DataFrame)
        self.assertEqual(
            set(df.index.names), {"classification_type", "global_metrics.field.name"}
        )
        self.assertEqual(
            set(df.columns), {"avg_f1_micro", "avg_nb_classes", "doc_count"}
        )

        self.assertEqual(
            df.to_dict(orient="index"),
            {
                ("multiclass", "gpc"): {
                    "avg_f1_micro": 0.93,
                    "avg_nb_classes": 211.12,
                    "doc_count": 198,
                },
                ("multiclass", "kind"): {
                    "avg_f1_micro": 0.89,
                    "avg_nb_classes": 206.5,
                    "doc_count": 370,
                },
                ("multilabel", "ispracticecompatible"): {
                    "avg_f1_micro": 0.72,
                    "avg_nb_classes": 18.71,
                    "doc_count": 128,
                },
                ("multilabel", "preservationmethods"): {
                    "avg_f1_micro": 0.8,
                    "avg_nb_classes": 9.97,
                    "doc_count": 76,
                },
            },
        )

    def test_grouping_agg(self):
        my_agg = Aggs(sample.EXPECTED_AGG_QUERY, mapping=MAPPING)
        agg_response = Aggregations(
            data=sample.ES_AGG_RESPONSE, search=Search().aggs(my_agg)
        )

        # none provided
        self.assertIsNone(agg_response._grouping_agg()[0])
        # fake provided
        with self.assertRaises(KeyError):
            agg_response._grouping_agg("yolo")
        # not bucket provided
        with self.assertRaises(ValueError):
            agg_response._grouping_agg("avg_f1_micro")
        # real provided
        self.assertEqual(
            agg_response._grouping_agg("global_metrics.field.name")[0],
            "global_metrics.field.name",
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from builtins import str as text
from six import text_type

import json

from future.utils import python_2_unicode_compatible

from pandagg.tree._tree import Tree
from pandagg.tree.mapping import _mapping

from pandagg.node.aggs.abstract import BucketAggNode, AggNode, Root
from pandagg.node.aggs.bucket import Nested, ReverseNested
from pandagg.node.aggs.pipeline import BucketSelector, BucketSort


@python_2_unicode_compatible
class Aggs(Tree):
    r"""
    Combination of aggregation clauses. This class provides handful methods to build an aggregation (see
    :func:`~pandagg.tree.aggs.Aggs.aggs` and :func:`~pandagg.tree.aggs.Aggs.groupby`), and is used as well
    to parse aggregations response in handy formats.

    Mapping declaration is optional, but doing so validates aggregation validity and automatically handles missing
    nested clauses.

    All following syntaxes are identical:

    From a dict:
    >>> Aggs({"per_user": {"terms": {"field": "user"}}})

    from an other Aggregation (simply a copy):
    >>> Aggs(Aggs({"per_user": {"terms": {"field": "user"}}}))

    :Keyword Arguments:
        * *mapping* (``dict`` or ``pandagg.tree.mapping.Mapping``) --
          Mapping of requested indice(s). Providing it will validate aggregations validity, and add required nested
          clauses if missing.

        * *nested_autocorrect* (``bool``) --
          In case of missing nested clauses in aggregation, if True, automatically add missing nested clauses, else
          raise error.

        * remaining kwargs:
          Used as body in aggregation
    """

    node_class = AggNode

    def __init__(
        self, aggs=None, mapping=None, nested_autocorrect=None, _groupby_ptr=None
    ):
        self.mapping = _mapping(mapping)
        self.nested_autocorrect = nested_autocorrect
        super(Aggs, self).__init__()

        # the root node of an aggregation is just the initial empty dict
        root_node = Root()
        self.insert_node(root_node)
        self._groupby_ptr = self.root if _groupby_ptr is None else _groupby_ptr

        if aggs is not None:
            self._insert_aggs(aggs, at_root=True)

    def __nonzero__(self):
        return bool(self.to_dict())

    __bool__ = __nonzero__

    def _insert_agg(self, name, node, insert_below=None, at_root=None, groupby=False):
        """
        Using flat declaration:
        >>> Aggs().agg("per_user", "terms", field="user")
        >>> Aggs().agg("per_user", {"terms": {"field": "user"}})

        Using DSL class:
        >>> from pandagg.aggs import Terms
        >>> Aggs().agg("per_user", Terms(field='user'))

        Agg node insertion, accepts following syntax:
        - name="per_user", type_or_agg="terms", field="user"
        - name="per_user", type_or_agg=Terms(field='user')
        - name="per_user", type_or_agg={"terms": {"field": "user"}}

        insert children aggs as well:
        - name="per_user", type_or_agg="terms", field="user", aggs={"avg_spent_time": {"avg": {"field": "spent_time"}}}
        - name="per_user", type_or_agg=Terms(field='user', aggs={"avg_spent_time": Avg(field="spent_time")})
        - name="per_user", type_or_agg={"field": "user", aggs: {"avg_spent_time": {"avg": {"field": "spent_time"}}}}
        """
        if insert_below and at_root:
            raise ValueError('Must define at most one of "insert_below" or "at_root".')
        if at_root:
            insert_below = self.root
        # based on last groupby pointer
        if insert_below is None:
            insert_below = self._groupby_ptr

        if not isinstance(name, text_type):
            raise ValueError('Agg "name" must be a str.')

        _children_aggs = node._children or {}

        if groupby:
            if _children_aggs:
                raise ValueError("Cannot group by multiple aggs at once.")
            subs = [self.drop_subtree(cid) for cid in self.children_ids(insert_below)]
            self.insert(node, key=name, parent_id=insert_below)
            for sub_key, sub_tree in subs:
                self.insert(sub_tree, key=sub_key, parent_id=node.identifier)
            # moving pointer when using groupby
            self._groupby_ptr = node.identifier
            return

        # in aggs mode, do not move pointer
        self.insert_node(node=node, key=name, parent_id=insert_below)
        for child_name, child in _children_aggs.items():
            child_node = self._translate_agg(child_name, child)
            self._insert_agg(
                name=child_name, node=child_node, insert_below=node.identifier
            )

    def _clone_init(self, deep=False):
        return Aggs(
            mapping=self.mapping.clone(deep=deep) if self.mapping is not None else None,
            nested_autocorrect=self.nested_autocorrect,
            _groupby_ptr=self._groupby_ptr,
        )

    def _is_eligible_grouping_node(self, nid):
        """Return whether node can be used as grouping node."""
        _, node = self.get(nid)
        if not isinstance(node, BucketAggNode):
            return False
        # special aggregations not returning anything
        if isinstance(node, (BucketSelector, BucketSort)):
            return False
        return True

    @property
    def deepest_linear_bucket_agg(self):
        """
        Return deepest bucket aggregation node (pandagg.nodes.abstract.BucketAggNode) of that aggregation that
        neither has siblings, nor has an ancestor with siblings.
        """
        if len(self._nodes_map) <= 1:
            return self.root
        last_bucket_agg_id = self.root
        children = [
            c
            for k, c in self.children(last_bucket_agg_id)
            if self._is_eligible_grouping_node(c.identifier)
        ]
        while len(children) == 1:
            last_agg_node = children[0]
            if not self._is_eligible_grouping_node(last_agg_node.identifier):
                break
            last_bucket_agg_id = last_agg_node.identifier
            children = [
                c
                for k, c in self.children(last_bucket_agg_id)
                if self._is_eligible_grouping_node(c.identifier)
            ]
        return last_bucket_agg_id

    def _validate_aggs_parent_id(self, pid):
        """
        If pid is not None, ensure that pid belongs to tree, and that it refers to a bucket aggregation.

        If pid not provided:

        - if previous groupby call, return deepest inserted node by last groupby call
        - else return deepest bucket aggregation if there is no ambiguity (linear aggregations).

        KO: non-ambiguous::
            A──> B──> C1
                 └──> C2
            raise error

        OK: non-ambiguous (linear)::

            A──> B──> C1
            return C1

        """
        if pid is not None:
            if not self._is_eligible_grouping_node(pid):
                raise ValueError("Node id <%s> is not a bucket aggregation." % pid)
            return pid
        if len(self._nodes_map) <= 1:
            return self.root

        # previous groupby calls defined grouping pointer
        if self._groupby_ptr is not None:
            return self._groupby_ptr

        eligible_nids = [
            n.identifier
            for _, n in self.list()
            if self._is_eligible_grouping_node(n.identifier)
        ]
        if len(eligible_nids) == 0:
            return self.root
        if len(eligible_nids) > 1:
            raise ValueError(
                "Declaration is ambiguous, you must declare the node id under which these "
                "aggregations should be placed."
            )
        return eligible_nids[0]

    def groupby(self, name, type_or_agg=None, insert_below=None, at_root=None, **body):
        r"""
        Arrange passed aggregations in vertical/nested manner, above or below another agg clause.

        Given the initial aggregation::

            A──> B
            └──> C

        If `insert_below` = 'A'::

            A──> new──> B
                  └──> C

        If `insert_above` = 'B'::

            A──> new──> B
            └──> C

        >>> Aggs()\
        >>> .groupby('per_user_id', 'terms', field='user_id')
        {"per_user_id":{"terms":{"field":"user_id"}}}

        >>> Aggs()\
        >>> .groupby('per_user_id', {'terms': {"field": "user_id"}})
        {"per_user_id":{"terms":{"field":"user_id"}}}

        >>> from pandagg.aggs import Terms
        >>> Aggs()\
        >>> .groupby('per_user_id', Terms(field="user_id"))
        {"per_user_id":{"terms":{"field":"user_id"}}}

        Accepted declarations for multiple aggregations:

        * *insert_below* (``string``) --
          Parent aggregation name under which these aggregations should be placed
        * *at_root* (``string``) --
          Insert aggregations at root of aggregation query

        * remaining kwargs:
          Used as body in aggregation

        :rtype: pandagg.aggs.Aggs
        """
        new_agg = self.clone(with_nodes=True)
        if insert_below is not None:
            insert_below = new_agg.id_from_key(insert_below)
        node = self._translate_agg(name, type_or_agg, **body)
        new_agg._insert_agg(
            name=name,
            node=node,
            insert_below=insert_below,
            at_root=at_root,
            groupby=True,
        )
        return new_agg

    def aggs(self, aggs, insert_below=None, at_root=False):
        r"""
        Arrange passed aggregations "horizontally".

        Given the initial aggregation::

            A──> B
            └──> C

        If passing multiple aggregations with `insert_below` = 'A'::

            A──> B
            └──> C
            └──> new1
            └──> new2

        Note: those will be placed under the `insert_below` aggregation clause id if provided, else under the deepest
        linear bucket aggregation if there is no ambiguity:

        OK::

            A──> B ─> C ─> new

        KO::

            A──> B
            └──> C

        `args` accepts single occurrence or sequence of following formats:

        * string (for terms agg concise declaration)
        * regular Elasticsearch dict syntax
        * AggNode instance (for instance Terms, Filters etc)


        :Keyword Arguments:
            * *insert_below* (``string``) --
              Parent aggregation name under which these aggregations should be placed
            * *at_root* (``string``) --
              Insert aggregations at root of aggregation query

            * remaining kwargs:
              Used as body in aggregation

        :rtype: pandagg.aggs.Aggs
        """
        new_agg = self.clone(with_nodes=True)
        if insert_below is not None:
            insert_below = new_agg.id_from_key(insert_below)
        new_agg._insert_aggs(aggs=aggs, insert_below=insert_below, at_root=at_root)
        return new_agg

    def _insert_aggs(self, aggs, insert_below=None, at_root=False):
        if at_root:
            insert_below = self.root
        elif not insert_below:
            # parent based on last groupby pointer
            if self._groupby_ptr:
                insert_below = self._groupby_ptr
            else:
                insert_below = self.deepest_linear_bucket_agg
        if isinstance(aggs, Aggs):
            self.merge(aggs, nid=insert_below)
            self._groupby_ptr = self.root
        elif isinstance(aggs, dict):
            for agg_name, agg_body in aggs.items():
                node = self._translate_agg(agg_name, agg_body)
                self._insert_agg(name=agg_name, node=node, insert_below=insert_below)
        elif aggs is not None:
            raise TypeError("Unsupported aggs type %s for Aggs" % type(aggs))

    @classmethod
    def _translate_agg(cls, name, type_or_agg=None, **body):
        """Accept multiple syntaxes, return a AggNode.
        :param type_or_agg:
        :param body:
        :return: AggNode
        """
        if isinstance(type_or_agg, text_type):
            # _translate_agg("per_user", "terms", field="user")
            return cls.get_node_dsl_class(type_or_agg)(**body)
        if isinstance(type_or_agg, AggNode):
            # _translate_agg("per_user", Terms(field='user'))
            if body:
                raise ValueError(
                    'Body cannot be added using "AggNode" declaration, got %s.' % body
                )
            return type_or_agg
        if isinstance(type_or_agg, dict):
            # _translate_agg("per_user", {"terms": {"field": "user"}})
            if body:
                raise ValueError(
                    'Body cannot be added using "dict" agg declaration, got %s.' % body
                )
            type_or_agg = type_or_agg.copy()
            children_aggs = (
                type_or_agg.pop("aggs", None)
                or type_or_agg.pop("aggregations", None)
                or {}
            )
            if len(type_or_agg) != 1:
                raise ValueError(
                    "Invalid aggregation declaration (two many keys): got <%s>"
                    % type_or_agg
                )
            type_, body_ = type_or_agg.popitem()
            body_ = body_.copy()
            if children_aggs:
                body_["aggs"] = children_aggs
            return cls.get_node_dsl_class(type_)(**body_)
        if type_or_agg is None:
            # if type_or_agg is not provided, by default execute a terms aggregation
            # _translate_agg("per_user")
            return cls.get_node_dsl_class("terms")(field=name, **body)
        raise ValueError('"type_or_agg" must be among "dict", "AggNode", "str"')

    def agg(self, name, type_or_agg=None, insert_below=None, at_root=False, **body):
        """

        :param name:
        :param type_or_agg:
        :param insert_below:
        :param at_root:
        :param body:
        :return:
        """
        new_agg = self.clone(with_nodes=True)
        if insert_below is not None:
            insert_below = new_agg.id_from_key(insert_below)
        node = self._translate_agg(name, type_or_agg, **body)
        new_agg._insert_agg(
            name=name, node=node, insert_below=insert_below, at_root=at_root
        )
        return new_agg

    def to_dict(self, from_=None, depth=None):
        if self.root is None:
            return None
        from_ = self.root if from_ is None else from_
        _, node = self.get(from_)
        children_queries = {}
        if depth is None or depth > 0:
            if depth is not None:
                depth -= 1
            for child_name, child_node in self.children(node.identifier):
                children_queries[child_name] = self.to_dict(
                    from_=child_node.identifier, depth=depth
                )
        if node.identifier == self.root:
            return children_queries
        node_query_dict = node.to_dict()
        if children_queries:
            node_query_dict["aggs"] = children_queries
        return node_query_dict

    def applied_nested_path_at_node(self, nid):
        # from current node to root
        for id_ in self.ancestors_ids(nid, include_current=True):
            _, node = self.get(id_)
            if isinstance(node, (Nested, ReverseNested)):
                return node.path
        return None

    def _insert_node_below(self, node, parent_id, key, by_path):
        """If mapping is provided, nested aggregations are automatically applied."""
        # if aggregation node is explicitely nested or reverse nested aggregation, do not override, but validate
        if (
            isinstance(node, Nested)
            or isinstance(node, ReverseNested)
            or not self.mapping
            or not hasattr(node, "field")
            or self.root is None
        ):
            return super(Aggs, self)._insert_node_below(node, parent_id, key, by_path)

        self.mapping.validate_agg_node(node)

        # from deepest to highest
        required_nested_level = self.mapping.nested_at_field(node.field)

        current_nested_level = self.applied_nested_path_at_node(parent_id)
        if current_nested_level == required_nested_level:
            return super(Aggs, self)._insert_node_below(node, parent_id, key, by_path)
        if not self.nested_autocorrect:
            raise ValueError(
                "Invalid %s agg on %s field. Invalid nested: expected %s, current %s."
                % (node.KEY, node.field, required_nested_level, current_nested_level)
            )
        if current_nested_level and (
            required_nested_level or "" in current_nested_level
        ):
            # requires reverse-nested
            # check if already exists in direct children, else create it
            child_reverse_nested = next(
                (
                    n
                    for k, n in self.children(parent_id)
                    if isinstance(n, ReverseNested) and n.path == required_nested_level
                ),
                None,
            )
            if child_reverse_nested:
                return super(Aggs, self)._insert_node_below(
                    node, child_reverse_nested.identifier, key, by_path
                )
            else:
                rv_node = ReverseNested()
                super(Aggs, self).insert_node(
                    rv_node,
                    parent_id=parent_id,
                    key="reverse_nested_below_%s" % self.get_key(parent_id),
                )
                return super(Aggs, self)._insert_node_below(
                    node, rv_node.identifier, key, by_path
                )

        # requires nested - apply all required nested fields
        for nested_lvl in reversed(self.mapping.list_nesteds_at_field(node.field)):
            if current_nested_level != nested_lvl:
                # check if already exists in direct children, else create it
                child_nested = next(
                    (
                        n
                        for k, n in (
                            self.children(parent_id) if parent_id is not None else []
                        )
                        if isinstance(n, Nested) and n.path == nested_lvl
                    ),
                    None,
                )
                if child_nested:
                    parent_id = child_nested.identifier
                    continue
                nested_node_name = (
                    "nested_below_root"
                    if parent_id is None
                    else "nested_below_%s" % self.get_key(parent_id)
                )
                nested_node = Nested(path=nested_lvl)
                super(Aggs, self)._insert_node_below(
                    nested_node, parent_id, nested_node_name, by_path
                )
                parent_id = nested_node.identifier
        super(Aggs, self)._insert_node_below(node, parent_id, key, by_path)

    def __str__(self):
        return json.dumps(self.to_dict(), indent=2)

    def show(self, *args, line_max_length=80, **kwargs):
        root_children = self.children(self.root)
        if len(root_children) == 0:
            return "<Aggregations> empty"
        if len(root_children) == 1:
            child_id = root_children[0][1].identifier
            return "<Aggregations>\n%s" % text(
                super(Tree, self).show(
                    child_id, *args, line_max_length=line_max_length, **kwargs
                )
            )

        return "<Aggregations>\n%s" % text(
            super(Tree, self).show(*args, line_max_length=line_max_length, **kwargs)
        )



from .abstract import LeafQueryClause, FlatFieldQueryClause, MultiFieldsQueryClause


class DistanceFeature(FlatFieldQueryClause):
    KEY = "distance_feature"


class MoreLikeThis(MultiFieldsQueryClause):
    KEY = "more_like_this"


class Percolate(FlatFieldQueryClause):
    KEY = "percolate"


class RankFeature(FlatFieldQueryClause):
    KEY = "rank_feature"


class Script(LeafQueryClause):
    KEY = "script"


class Wrapper(LeafQueryClause):
    KEY = "wrapper"



from mock import Mock, patch

from pandagg import Search
from pandagg.tree.aggs.aggs import Aggs
from pandagg.tree.response import AggsResponseTree
from pandagg.interactive.response import IResponse
from tests import PandaggTestCase
from tests.testing_samples.mapping_example import MAPPING
import tests.testing_samples.data_sample as sample


class ClientBoundResponseTestCase(PandaggTestCase):
    @patch("lighttree.node.uuid.uuid4")
    def test_client_bound_response(self, uuid_mock):
        uuid_mock.side_effect = range(1000)
        client_mock = Mock(spec=["search"])

        my_agg = Aggs(sample.EXPECTED_AGG_QUERY, mapping=MAPPING)
        response_tree = AggsResponseTree(aggs=my_agg).parse(sample.ES_AGG_RESPONSE)

        response = IResponse(
            search=Search(index="some_index", using=client_mock).query(
                {"term": {"some_field": 1}}
            ),
            tree=response_tree,
            depth=1,
        )

        # ensure that navigation to attributes works with autocompletion (dir is used in ipython)
        self.assertIn("classification_type_multiclass", dir(response))
        self.assertIn("classification_type_multilabel", dir(response))

        multiclass = response.classification_type_multiclass
        self.assertIsInstance(multiclass, IResponse)
        self.assertIs(multiclass._initial_tree, response._tree)

        self.assertIn("global_metrics_field_name_gpc", dir(multiclass))
        gpc = multiclass.global_metrics_field_name_gpc
        self.assertIsInstance(gpc, IResponse)
        self.assertIs(gpc._initial_tree, response._tree)

        # test filter query used to list documents belonging to bucket
        self.assertQueryEqual(
            gpc.get_bucket_filter(),
            {
                "bool": {
                    "must": [
                        {"term": {"global_metrics.field.name": {"value": "gpc"}}},
                        {"term": {"classification_type": {"value": "multiclass"}}},
                    ]
                }
            },
        )

        # convert to search request (includes initial query clauses)
        srequest = gpc.search()
        self.assertIsInstance(srequest, Search)
        self.assertSearchEqual(
            srequest.to_dict(),
            {
                "query": {
                    "bool": {
                        "must": [
                            {"term": {"global_metrics.field.name": {"value": "gpc"}}},
                            {"term": {"classification_type": {"value": "multiclass"}}},
                            {"term": {"some_field": {"value": 1}}},
                        ]
                    }
                }
            },
        )
        self.assertEqual(srequest._aggs.to_dict(), {})



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from pandagg.node._node import Node


class BucketNode(Node):
    def __init__(self):
        self.level = None
        super(BucketNode, self).__init__(keyed=False)


class Bucket(BucketNode):
    def __init__(self, value, key=None, level=None):
        super(Bucket, self).__init__()
        self.value = value
        self.level = level
        self.key = key

    @property
    def attr_name(self):
        """Determine under which attribute name the bucket will be available in response tree.
        Dots are replaced by `_` characters so that they don't prevent from accessing as attribute.

        Resulting attribute unfit for python attribute name syntax is still possible and will be accessible through
        item access (dict like), see more in 'utils.Obj' for more details.
        """
        if self.key is not None:
            return "%s_%s" % (self.level.replace(".", "_"), self._coerced_key)
        return self.level.replace(".", "_")

    def line_repr(self, **kwargs):
        s = self.level or ""
        if self.key is not None:
            s += "=%s" % self._coerced_key
        return s, str(self.value) if self.value else ""

    @property
    def _coerced_key(self):
        key = self.key
        try:
            # order matters, will
            key = float(key)
            key = int(key)
        except (ValueError, TypeError):
            pass
        return key



from pandagg.node.query.abstract import QueryClause


class CompoundClause(QueryClause):
    """Compound clauses can encapsulate other query clauses:

    .. code-block::

        {
            "<query_type>" : {
                <query_body>
                <children_clauses>
            }
        }

    """

    _default_operator = None
    # key -> multiple / single
    _parent_params = {}

    def __init__(self, _name=None, **body):
        b = body.copy()
        children = {}
        for param in self._parent_params:
            v = b.pop(param, None)
            if not v:
                continue
            children[param] = v if isinstance(v, list) else [v]
        super(CompoundClause, self).__init__(
            _name=_name, accept_children=True, keyed=True, _children=children, **b
        )


class Bool(CompoundClause):
    """
    >>> Bool(must=[], should=[], filter=[], must_not=[], boost=1.2)
    """

    _default_operator = "must"
    _parent_params = {"should": True, "must": True, "must_not": True, "filter": True}
    KEY = "bool"


class Boosting(CompoundClause):
    _default_operator = "positive"
    _parent_params = ["positive", "negative"]
    KEY = "boosting"


class ConstantScore(CompoundClause):
    _default_operator = "filter"
    _parent_params = ["filter", "boost"]
    KEY = "constant_score"


class DisMax(CompoundClause):
    _default_operator = "queries"
    _parent_params = ["queries"]
    KEY = "dis_max"


class FunctionScore(CompoundClause):
    _default_operator = "query"
    _parent_params = ["query"]
    KEY = "function_score"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from future.utils import python_2_unicode_compatible

from builtins import str as text

from lighttree import Tree as OriginalTree

from pandagg.utils import DSLMixin


@python_2_unicode_compatible
class Tree(DSLMixin, OriginalTree):

    KEY = None
    _type_name = None

    @classmethod
    def get_node_dsl_class(cls, name):
        return cls.node_class._get_dsl_class(name)

    def id_from_key(self, key):
        """Find node identifier based on key. If multiple nodes have the same key, takes the first found one."""
        for k, n in self.list():
            if k == key:
                return n.identifier
        raise KeyError('No node found with key "%s"' % key)

    def __str__(self):
        return "<{class_}>\n{tree}".format(
            class_=text(self.__class__.__name__), tree=self.show(limit=40)
        )

    def __repr__(self):
        return self.__str__()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import json

from future.utils import python_2_unicode_compatible, text_type
from builtins import str as text

from pandagg._decorators import Substitution
from pandagg.node.query._parameter_clause import ParentParameterClause
from pandagg.node.query.abstract import QueryClause, LeafQueryClause
from pandagg.node.query.compound import CompoundClause, Bool
from pandagg.node.query.joining import Nested

from pandagg.tree._tree import Tree
from pandagg.tree.mapping import _mapping

ADD = "add"
REPLACE = "replace"
REPLACE_ALL = "replace_all"

sub_insertion = Substitution(
    insertion_doc="""
    * *parent* (``str``) --
      named query clause under which the inserted clauses should be placed.

    * *parent_param* (``str`` optional parameter when using *parent* param) --
      parameter under which inserted clauses will be placed. For instance if *parent* clause is a boolean, can be
      'must', 'filter', 'should', 'must_not'.

    * *child* (``str``) --
      named query clause above which the inserted clauses should be placed.

    * *child_param* (``str`` optional parameter when using *parent* param) --
      parameter of inserted boolean clause under which child clauses will be placed. For instance if inserted clause
      is a boolean, can be 'must', 'filter', 'should', 'must_not'.

    * *mode* (``str`` one of 'add', 'replace', 'replace_all') --
      merging strategy when inserting clauses on a existing compound clause.

      - 'add' (default) : adds new clauses keeping initial ones
      - 'replace' : for each parameter (for instance in 'bool' case : 'filter', 'must', 'must_not', 'should'),
        replace existing clauses under this parameter, by new ones only if declared in inserted compound query
      - 'replace_all' : existing compound clause is completely replaced by the new one
"""
)


@python_2_unicode_compatible
class Query(Tree):
    r"""Combination of query clauses.

    Mapping declaration is optional, but doing so validates query validity and automatically inserts nested clauses
    when necessary.

    :Keyword Arguments:
        * *mapping* (``dict`` or ``pandagg.tree.mapping.Mapping``) --
          Mapping of requested indice(s). Providing it will add validation features, and add required nested
          clauses if missing.

        * *nested_autocorrect* (``bool``) --
          In case of missing nested clauses in query, if True, automatically add missing nested clauses, else raise
          error.

        * remaining kwargs:
          Used as body in query clauses.
    """

    KEY = None
    node_class = QueryClause

    def __init__(self, q=None, mapping=None, nested_autocorrect=False):
        self.mapping = _mapping(mapping)
        self.nested_autocorrect = nested_autocorrect
        super(Query, self).__init__()
        if q:
            self._insert_query(q)

    def __nonzero__(self):
        return bool(self.to_dict())

    __bool__ = __nonzero__

    def _clone_init(self, deep=False):
        return Query(
            mapping=None
            if self.mapping is None
            else self.mapping.clone(with_nodes=True, deep=deep),
            nested_autocorrect=self.nested_autocorrect,
        )

    @classmethod
    def _from_dict(cls, d):
        """return Query, from dict"""
        # {"nested": {"path": "xxx", "query": {"term": {"some_field": "234"}}}}
        if not len(d.keys()) == 1:
            raise ValueError("Wrong declaration, got %s" % d.keys())
        vk, vv = d.copy().popitem()
        return cls._get_dsl_class_from_tree_or_node(vk, **vv)

    def _has_bool_root(self):
        if not self.root:
            return False
        _, r = self.get(self.root)
        return isinstance(r, Bool)

    def _compound_param_id(self, nid, key, create_if_not_exists=True):
        """
        :param nid: id of compound node
        :param key: param key, for instance if compound if bool, can be 'must', 'should', 'must_not' etc
        :return: param node id
        """
        try:
            return self.child_id(nid, key)
        except ValueError:
            if not create_if_not_exists:
                raise
        # add key
        param_node = self.get_node_dsl_class(key)()
        self._insert_node_below(param_node, parent_id=nid, key=key, by_path=False)
        return param_node.identifier

    @classmethod
    def _translate_query(cls, type_or_query=None, **body):
        """Accept multiple syntaxes, return a QueryClause node.
        :param type_or_query:
        :param body:
        :return: QueryClause
        """
        if isinstance(type_or_query, QueryClause):
            if body:
                raise ValueError(
                    'Body cannot be added using "QueryClause" declaration, got %s.'
                    % body
                )
            return type_or_query
        if isinstance(type_or_query, Query):
            type_or_query = type_or_query.to_dict()
        if isinstance(type_or_query, dict):
            if body:
                raise ValueError(
                    'Body cannot be added using "dict" query clause declaration, got %s.'
                    % body
                )
            type_or_query = type_or_query.copy()
            # {"term": {"some_field": 1}}
            # {"bool": {"filter": [{"term": {"some_field": 1}}]}}
            if len(type_or_query) != 1:
                raise ValueError(
                    "Invalid query clause declaration (two many keys): got <%s>"
                    % type_or_query
                )
            type_, body_ = type_or_query.popitem()
            return cls.get_node_dsl_class(type_)(**body_)
        if isinstance(type_or_query, text_type):
            return cls.get_node_dsl_class(type_or_query)(**body)
        raise ValueError('"type_or_query" must be among "dict", "AggNode", "str"')

    def _insert_query_at(
        self, node, mode, on=None, insert_below=None, compound_param=None
    ):
        """Insert clause (and its children) in query.

        If compound query with on specified: merge according to mode.
        If insert_below is not specified, place on top (wrapped in bool-must if necessary).
        If insert_below is provided (only under compound query): place under it.

        :param node: node to insert, can contain children clauses (in case of compound query).
        :param mode: how compound queries merge should be treated
        :param on: id of compound query on which node should be merged
        :param insert_below:
        :return:
        """
        if mode not in (ADD, REPLACE, REPLACE_ALL):
            raise ValueError("Invalid mode %s" % mode)

        if (
            isinstance(node, Bool)
            and not on
            and not insert_below
            and self._has_bool_root()
        ):
            on = self.root

        if isinstance(node, CompoundClause) and on:
            # ensure we try to merge on same type of clause
            _, existing = self.get(on)
            if existing.KEY != node.KEY:
                raise ValueError(
                    "Cannot merge compound clause %s on %s. Must be the same."
                    % (node.KEY, existing.KEY)
                )
            if mode == REPLACE_ALL:
                pid = self.parent_id(on)
                existing_k, _ = self.drop_subtree(on)
                self._insert_query(node, insert_below=pid)
                existing.body = node.body
                return

            # merge
            existing.body.update(node.body)
            for param_key, children in node._children.items():
                if not children:
                    continue
                param_id = self._compound_param_id(on, param_key)
                # here, possible modes are either ADD, or REPLACE
                if mode == REPLACE:
                    existing_clauses_ids = self.children_ids(param_id)
                    for eid in existing_clauses_ids:
                        self.drop_node(eid)
                for child in children:
                    self._insert_query(child, insert_below=param_id)
            return

        if insert_below:
            # below node, with compound_param
            _, pnode = self.get(insert_below)
            if not isinstance(pnode, CompoundClause):
                raise ValueError(
                    "Cannot insert clause below %s clause (only compound clauses can have children clauses)."
                    % pnode.KEY
                )
            compound_param = compound_param or pnode._default_operator
            if compound_param not in pnode._parent_params.keys():
                raise ValueError(
                    "<%s> parameter for <%s> compound clause does not accept children clauses."
                    % (compound_param, pnode.KEY)
                )
            param_id = self._compound_param_id(insert_below, compound_param)
            _, p = self.get(param_id)
            if not p.MULTIPLE:
                # inserting a clause, under a parameter allowing a single clause (for instance below nested query)
                cids = self.children_ids(param_id)
                if cids:
                    # can be at most one
                    cid = cids[0]
                    # must place existing clause, and new one under a bool -> must
                    _, existing_child = self.get(cid)
                    if isinstance(existing_child, Bool):
                        child_param_id = self._compound_param_id(
                            existing_child.identifier, "must"
                        )
                        self._insert_query(node, insert_below=child_param_id)
                        return
                    _, existing_child = self.drop_node(cid)
                    self._insert_query(
                        Bool(must=[existing_child, node]), insert_below=param_id
                    )
                    return
            self._insert_query(node, insert_below=param_id)
            return

        # from now on: position was not provided:
        if self.is_empty():
            # currently empty (-> on top)
            self._insert_query(node)
            return

        if self._has_bool_root():
            # top query is bool
            must_id = self._compound_param_id(self.root, "must")
            self._insert_query(node, insert_below=must_id)
            return

        # top query is not bool
        _, initial_query = self.drop_subtree(self.root)
        if isinstance(node, Bool):
            # if inserted node is bool, placed on top, and previous query is added under "must" parameter
            self._insert_query(node)
            self._insert_query_at(
                initial_query.to_dict(), insert_below=node.identifier, mode=ADD
            )
        else:
            # we place initial + added clauses under a bool>must
            self._insert_query(Bool(must=[node, initial_query.to_dict()]))
        return

    def _insert_query(self, query=None, insert_below=None):
        """
        Accept Node, or dict syntaxes, convert to nodes.
        Insert query clause and its children.
        Wraps in bool->must if parent param is not multiple.

        Does not handle:
        - syntax conversion (Tree -> Node, or flat syntax)
        - logic about where to insert it, should be handled before (dumb insert below insert_below)

        >>> Query()._insert_query({"terms": {"field": "user"}})
        >>> Query()._insert_query(Query({"terms": {"field": "user"}}))
        >>> Query()._insert_query({"bool": {"must": {"term": {"some_field": "yolo"}}, "should": {}}})
        >>> Query()._insert_query({"term": {"some_field": "yolo"})
        """
        if isinstance(query, QueryClause):
            node = query
        elif isinstance(query, dict):
            query = query.copy()
            # {"term": {"some_field": 1}}
            # {"bool": {"filter": [{"term": {"some_field": 1}}]}}
            if len(query.keys()) != 1:
                raise ValueError(
                    "Invalid query clause declaration (two many keys): got <%s>"
                    % query.keys()
                )
            type_, body_ = query.popitem()
            node_klass = self.get_node_dsl_class(type_)
            if issubclass(node_klass, ParentParameterClause):
                raise ValueError()
            node = node_klass(**body_)
        else:
            raise ValueError('"query" must be of type "dict" or "AggNode"')

        self.insert_node(node, parent_id=insert_below)

        if not isinstance(node, CompoundClause):
            return

        _children_clauses = node._children.copy()
        for param_name, child_nodes in _children_clauses.items():
            param_node = self.get_node_dsl_class(param_name)()
            if not param_node.MULTIPLE and len(child_nodes) > 1:
                raise ValueError(
                    "Cannot insert multiple query clauses under %s parameter"
                    % param_name
                )
            self.insert_node(param_node, parent_id=node.identifier, key=param_name)
            for child in child_nodes:
                self._insert_query(query=child, insert_below=param_node.identifier)

    def _insert_node_below(self, node, parent_id, key, by_path):
        """Override lighttree.Tree._insert_node_below method to ensure inserted query clause is consistent."""
        if parent_id is not None:
            _, pnode = self.get(parent_id)
            if isinstance(pnode, LeafQueryClause):
                raise ValueError(
                    "Cannot add clause under leaf query clause <%s>" % pnode.KEY
                )
            if isinstance(pnode, ParentParameterClause):
                if isinstance(node, ParentParameterClause):
                    raise ValueError(
                        "Cannot add parameter clause <%s> under another parameter clause <%s>"
                        % (pnode.KEY, node.KEY)
                    )
            if isinstance(pnode, CompoundClause):
                if key not in pnode._parent_params:
                    raise ValueError(
                        "Expect a parameter clause of type %s under <%s> compound clause, got <%s>"
                        % (pnode._parent_params.keys(), pnode.KEY, key)
                    )

        # automatic handling of nested clauses
        if isinstance(node, Nested) or not self.mapping or not hasattr(node, "field"):
            return super(Query, self)._insert_node_below(
                node=node, parent_id=parent_id, key=key, by_path=by_path
            )
        required_nested_level = self.mapping.nested_at_field(node.field)
        if len(self.list()) <= 1:
            # empty
            current_nested_level = None
        else:
            current_nested_level = self.applied_nested_path_at_node(parent_id)
        if current_nested_level == required_nested_level:
            return super(Query, self)._insert_node_below(
                node=node, parent_id=parent_id, key=key, by_path=by_path
            )
        if not self.nested_autocorrect:
            raise ValueError(
                "Invalid %s query clause on %s field. Invalid nested: expected %s, current %s."
                % (node.KEY, node.field, required_nested_level, current_nested_level)
            )
        # requires nested - apply all required nested fields
        to_insert = node
        for nested_lvl in self.mapping.list_nesteds_at_field(node.field):
            if current_nested_level != nested_lvl:
                to_insert = self.get_node_dsl_class("nested")(
                    path=nested_lvl, query=to_insert
                )
        self._insert_query(to_insert, parent_id)

    def applied_nested_path_at_node(self, nid):
        # from current node to root
        for id_ in self.ancestors_ids(nid, include_current=True):
            _, node = self.get(id_)
            if isinstance(node, Nested):
                return node.path
        return None

    def to_dict(self, from_=None):
        """Serialize query as dict."""
        if self.root is None:
            return None
        from_ = self.root if from_ is None else from_
        key, node = self.get(from_)
        if isinstance(node, LeafQueryClause):
            return node.to_dict()

        if not isinstance(node, CompoundClause):
            raise ValueError("Unexpected %s" % node.__class__)

        d = {}
        is_empty = True
        for param_key, param_node in self.children(node.identifier):
            children_serialized = [
                self.to_dict(child_node.identifier)
                for _, child_node in self.children(param_node.identifier)
            ]
            children_serialized = [c for c in children_serialized if c]
            if not children_serialized:
                continue
            is_empty = False
            if not param_node.MULTIPLE:
                d[param_key] = children_serialized[0]
                continue
            d[param_key] = children_serialized
        # a compound query clause can exist, without children clauses, in that case, just ignore it
        if is_empty:
            return None
        q = node.to_dict()
        q[node.KEY].update(d)
        return q

    @sub_insertion
    def query(
        self,
        type_or_query,
        insert_below=None,
        on=None,
        mode=ADD,
        compound_param=None,
        **body
    ):
        r"""Insert new clause(s) in current query.

        Inserted clause can accepts following syntaxes.

        Given an empty query:

        >>> from pandagg.query import Query
        >>> q = Query()

        flat syntax: clause type, followed by query clause body as keyword arguments:

        >>> q.query('term', some_field=23)
        {'term': {'some_field': 23}}

        using pandagg DSL:

        >>> from pandagg.query import Term
        >>> q.query(Term(field=23))
        {'term': {'some_field': 23}}

        >>> q.query({'bool': {'must': [{'term': {'some_field': 1}}]}})
        {'bool': {'must': [{'term': {'some_field': 1}}]}}

        >>> from pandagg.query import Bool
        >>> q.query(Bool(must=[{'term': {'some_field': 1}}], boost=1))

        :Keyword Arguments:
        %(insertion_doc)s

        """
        q = self.clone(with_nodes=True)
        node = self._translate_query(type_or_query, **body)
        q._insert_query_at(
            node,
            mode=mode,
            on=on,
            insert_below=insert_below,
            compound_param=compound_param,
        )
        return q

    # compound
    def bool(
        self,
        must=None,
        should=None,
        must_not=None,
        filter=None,
        insert_below=None,
        on=None,
        mode=ADD,
        **body
    ):
        """
        >>> Query().bool(must={"term": {"some_field": "yolo"}})
        """
        return self.query(
            "bool",
            must=must,
            should=should,
            must_not=must_not,
            filter=filter,
            insert_below=insert_below,
            on=on,
            mode=mode,
            **body
        )

    def boosting(
        self, positive=None, negative=None, insert_below=None, on=None, mode=ADD, **body
    ):
        if not positive and not negative:
            raise ValueError('Expect at least one of "positive", "negative"')
        return self.query(
            "boosting",
            positive=positive,
            negative=negative,
            insert_below=insert_below,
            on=on,
            mode=mode,
            **body
        )

    def constant_score(
        self, filter=None, boost=None, insert_below=None, on=None, mode=ADD, **body
    ):
        if not filter and not boost:
            raise ValueError('Expect at least one of "filter", "boost"')
        return self.query(
            "constant_score",
            filter=filter,
            boost=boost,
            insert_below=insert_below,
            on=on,
            mode=mode,
            **body
        )

    def dis_max(self, queries, insert_below=None, on=None, mode=ADD, **body):
        return self.query(
            "dis_max",
            queries=queries,
            insert_below=insert_below,
            on=on,
            mode=mode,
            **body
        )

    def function_score(self, query, insert_below=None, on=None, mode=ADD, **body):
        return self.query(
            "function_score",
            query=query,
            insert_below=insert_below,
            on=on,
            mode=mode,
            **body
        )

    def nested(self, path, query=None, insert_below=None, on=None, mode=ADD, **body):
        return self.query(
            "nested",
            query=query,
            insert_below=insert_below,
            on=on,
            mode=mode,
            path=path,
            **body
        )

    def has_child(self, query, insert_below=None, on=None, mode=ADD, **body):
        return self.query(
            "has_child",
            query=query,
            insert_below=insert_below,
            on=on,
            mode=mode,
            **body
        )

    def has_parent(self, query, insert_below=None, on=None, mode=ADD, **body):
        return self.query(
            "has_parent",
            query=query,
            insert_below=insert_below,
            on=on,
            mode=mode,
            **body
        )

    def script_score(self, query, insert_below=None, on=None, mode=ADD, **body):
        return self.query(
            "script_score",
            query=query,
            insert_below=insert_below,
            on=on,
            mode=mode,
            **body
        )

    def pinned_query(self, organic, insert_below=None, on=None, mode=ADD, **body):
        return self.query(
            "pinned_query",
            organic=organic,
            insert_below=insert_below,
            on=on,
            mode=mode,
            **body
        )

    # compound parameters
    def _compound_param_insert(
        self,
        compound_key,
        compound_param_key,
        mode,
        type_or_query,
        insert_below=None,
        on=None,
        compound_body=None,
        **body
    ):
        q = self.clone(with_nodes=True)
        node = self._translate_query(type_or_query, **body)
        compound_body = compound_body or {}
        compound_body[compound_param_key] = node
        compound_node = self.get_node_dsl_class(compound_key)(**compound_body)
        q._insert_query_at(compound_node, on=on, insert_below=insert_below, mode=mode)
        return q

    def must(
        self,
        type_or_query,
        insert_below=None,
        on=None,
        mode=ADD,
        bool_body=None,
        **body
    ):
        """
        >>> Query().must('term', some_field=1)
        >>> Query().must({'term': {'some_field': 1}})
        """
        return self._compound_param_insert(
            "bool", "must", mode, type_or_query, insert_below, on, bool_body, **body
        )

    def should(
        self,
        type_or_query,
        insert_below=None,
        on=None,
        mode=ADD,
        bool_body=None,
        **body
    ):
        return self._compound_param_insert(
            "bool", "should", mode, type_or_query, insert_below, on, bool_body, **body
        )

    def must_not(
        self,
        type_or_query,
        insert_below=None,
        on=None,
        mode=ADD,
        bool_body=None,
        **body
    ):
        return self._compound_param_insert(
            "bool", "must_not", mode, type_or_query, insert_below, on, bool_body, **body
        )

    def filter(
        self,
        type_or_query,
        insert_below=None,
        on=None,
        mode=ADD,
        bool_body=None,
        **body
    ):
        return self._compound_param_insert(
            "bool", "filter", mode, type_or_query, insert_below, on, bool_body, **body
        )

    def show(self, *args, line_max_length=80, **kwargs):
        return "<Query>\n%s" % text(
            super(Tree, self).show(*args, line_max_length=line_max_length, **kwargs)
        )

    def __str__(self):
        return json.dumps(self.to_dict(), indent=2)



from __future__ import unicode_literals

from unittest import TestCase

from pandagg.node.query.specialized import Wrapper
from pandagg.node.query import (
    DistanceFeature,
    MoreLikeThis,
    Percolate,
    RankFeature,
    Script,
)


class SpecializedQueriesTestCase(TestCase):
    def test_distance_feature_clause(self):
        body = {"field": "production_date", "pivot": "7d", "origin": "now"}
        expected = {"distance_feature": body}

        q = DistanceFeature(field="production_date", pivot="7d", origin="now")
        self.assertEqual(q.body, body)
        self.assertEqual(q.to_dict(), expected)
        self.assertEqual(
            q.line_repr(depth=None),
            ("distance_feature", 'field="production_date", origin="now", pivot="7d"'),
        )

    def test_more_like_this_clause(self):
        body = {
            "fields": ["title", "description"],
            "like": "Once upon a time",
            "min_term_freq": 1,
            "max_query_terms": 12,
        }
        expected = {"more_like_this": body}

        q = MoreLikeThis(
            fields=["title", "description"],
            like="Once upon a time",
            min_term_freq=1,
            max_query_terms=12,
        )
        self.assertEqual(q.body, body)
        self.assertEqual(q.to_dict(), expected)
        self.assertEqual(
            q.line_repr(depth=None),
            ("more_like_this", "fields=['title', 'description']"),
        )

    def test_percolate_clause(self):
        body = {
            "field": "query",
            "document": {"message": "A new bonsai tree in the office"},
        }
        expected = {"percolate": body}

        q = Percolate(
            field="query", document={"message": "A new bonsai tree in the office"}
        )
        self.assertEqual(q.body, body)
        self.assertEqual(q.to_dict(), expected)
        self.assertEqual(
            q.line_repr(depth=None),
            (
                "percolate",
                'document={"message": "A new bonsai tree in the office"}, field="query"',
            ),
        )

    def test_rank_feature_clause(self):
        body = {"field": "url_length", "boost": 0.1}
        expected = {"rank_feature": body}

        q = RankFeature(field="url_length", boost=0.1)
        self.assertEqual(q.body, body)
        self.assertEqual(q.to_dict(), expected)
        self.assertEqual(
            q.line_repr(depth=None), ("rank_feature", 'boost=0.1, field="url_length"')
        )

    def test_script_clause(self):
        body = {
            "script": {
                "source": "doc['num1'].value > params.param1",
                "lang": "painless",
                "params": {"param1": 5},
            }
        }
        expected = {"script": body}

        q = Script(
            script={
                "source": "doc['num1'].value > params.param1",
                "lang": "painless",
                "params": {"param1": 5},
            }
        )
        self.assertEqual(q.body, body)
        self.assertEqual(q.to_dict(), expected)
        self.assertEqual(
            q.line_repr(depth=None),
            (
                "script",
                'script={"lang": "painless", "params": {"param1": 5}, "source": '
                "\"doc['num1'].value > params.param1\"}",
            ),
        )

    def test_wrapper_clause(self):
        body = {"query": "eyJ0ZXJtIiA6IHsgInVzZXIiIDogIktpbWNoeSIgfX0="}
        expected = {"wrapper": body}

        q = Wrapper(query="eyJ0ZXJtIiA6IHsgInVzZXIiIDogIktpbWNoeSIgfX0=")
        self.assertEqual(q.body, body)
        self.assertEqual(q.to_dict(), expected)
        self.assertEqual(
            q.line_repr(depth=None),
            ("wrapper", 'query="eyJ0ZXJtIiA6IHsgInVzZXIiIDogIktpbWNoeSIgfX0="'),
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

NUMERIC_TYPES = [
    "long",
    "integer",
    "short",
    "byte",
    "double",
    "float",
    "half_float",
    "scaled_float",
    "ip",
    "token_count",
    "date",
    "boolean",
]

MAPPING_TYPES = [
    "binary",
    "geo_point",
    "geo_shape",
    "nested",
    "object",
    "text",
    "keyword",
    "flattened",
] + NUMERIC_TYPES



from __future__ import unicode_literals

from unittest import TestCase

from pandagg.node.query import (
    Intervals,
    Match,
    MatchBoolPrefix,
    MatchPhrase,
    MatchPhrasePrefix,
    MultiMatch,
    QueryString,
    SimpleQueryString,
)


class FullTextQueriesTestCase(TestCase):
    def test_interval_clause(self):
        body = {
            "some_field": {
                "all_of": {
                    "intervals": [
                        {"match": {"query": "the"}},
                        {
                            "any_of": {
                                "intervals": [
                                    {"match": {"query": "big"}},
                                    {"match": {"query": "big bad"}},
                                ]
                            }
                        },
                        {"match": {"query": "wolf"}},
                    ],
                    "max_gaps": 0,
                    "ordered": True,
                }
            }
        }
        expected = {"intervals": body}

        q1 = Intervals(
            field="some_field",
            all_of={
                "intervals": [
                    {"match": {"query": "the"}},
                    {
                        "any_of": {
                            "intervals": [
                                {"match": {"query": "big"}},
                                {"match": {"query": "big bad"}},
                            ]
                        }
                    },
                    {"match": {"query": "wolf"}},
                ],
                "max_gaps": 0,
                "ordered": True,
            },
        )
        q2 = Intervals(
            some_field={
                "all_of": {
                    "intervals": [
                        {"match": {"query": "the"}},
                        {
                            "any_of": {
                                "intervals": [
                                    {"match": {"query": "big"}},
                                    {"match": {"query": "big bad"}},
                                ]
                            }
                        },
                        {"match": {"query": "wolf"}},
                    ],
                    "max_gaps": 0,
                    "ordered": True,
                }
            }
        )
        for q in (q1, q2):
            self.assertEqual(q.body, body)
            self.assertEqual(q.to_dict(), expected)
            self.assertEqual(
                q.line_repr(depth=None),
                (
                    "intervals",
                    'field=some_field, all_of={"intervals": [{"match": {"query": "the"}}, '
                    '{"any_of": {"intervals": [{"match": {"query": "big"}}, {"match": {"query": '
                    '"big bad"}}]}}, {"match": {"query": "wolf"}}], "max_gaps": 0, "ordered": '
                    "true}",
                ),
            )

    def test_match_clause(self):
        body = {"message": {"query": "this is a test", "operator": "and"}}
        expected = {"match": body}

        q1 = Match(field="message", query="this is a test", operator="and")
        q2 = Match(message={"query": "this is a test", "operator": "and"})
        for q in (q1, q2):
            self.assertEqual(q.body, body)
            self.assertEqual(q.to_dict(), expected)
            self.assertEqual(
                q.line_repr(depth=None),
                ("match", 'field=message, operator="and", query="this is a test"'),
            )

        # short syntax
        q3 = Match(message="this is a test")
        self.assertEqual(q3.body, {"message": {"query": "this is a test"}})
        self.assertEqual(
            q3.to_dict(), {"match": {"message": {"query": "this is a test"}}}
        )
        self.assertEqual(
            q3.line_repr(depth=None), ("match", 'field=message, query="this is a test"')
        )

    def test_match_bool_prefix_clause(self):
        body = {"message": {"query": "quick brown f", "analyzer": "keyword"}}
        expected = {"match_bool_prefix": body}

        q1 = MatchBoolPrefix(field="message", query="quick brown f", analyzer="keyword")
        q2 = MatchBoolPrefix(message={"query": "quick brown f", "analyzer": "keyword"})
        for q in (q1, q2):
            self.assertEqual(q.body, body)
            self.assertEqual(q.to_dict(), expected)
            self.assertEqual(
                q.line_repr(depth=None),
                (
                    "match_bool_prefix",
                    'field=message, analyzer="keyword", query="quick brown f"',
                ),
            )

        # short syntax
        q3 = MatchBoolPrefix(message="quick brown f")
        self.assertEqual(q3.body, {"message": {"query": "quick brown f"}})
        self.assertEqual(
            q3.to_dict(), {"match_bool_prefix": {"message": {"query": "quick brown f"}}}
        )
        self.assertEqual(
            q3.line_repr(depth=None),
            ("match_bool_prefix", 'field=message, query="quick brown f"'),
        )

    def test_match_phrase_clause(self):
        body = {"message": {"query": "this is a test", "analyzer": "my_analyzer"}}
        expected = {"match_phrase": body}

        q1 = MatchPhrase(
            field="message", query="this is a test", analyzer="my_analyzer"
        )
        q2 = MatchPhrase(message={"query": "this is a test", "analyzer": "my_analyzer"})
        for q in (q1, q2):
            self.assertEqual(q.body, body)
            self.assertEqual(q.to_dict(), expected)
            self.assertEqual(
                q.line_repr(depth=None),
                (
                    "match_phrase",
                    'field=message, analyzer="my_analyzer", query="this is a test"',
                ),
            )

        # short syntax
        q3 = MatchPhrase(message="this is a test")
        self.assertEqual(q3.body, {"message": {"query": "this is a test"}})
        self.assertEqual(
            q3.to_dict(), {"match_phrase": {"message": {"query": "this is a test"}}}
        )
        self.assertEqual(
            q3.line_repr(depth=None),
            ("match_phrase", 'field=message, query="this is a test"'),
        )

    def test_match_phrase_prefix_clause(self):
        body = {"message": {"query": "this is a test", "analyzer": "my_analyzer"}}
        expected = {"match_phrase_prefix": body}

        q1 = MatchPhrasePrefix(
            field="message", query="this is a test", analyzer="my_analyzer"
        )
        q2 = MatchPhrasePrefix(
            message={"query": "this is a test", "analyzer": "my_analyzer"}
        )
        for q in (q1, q2):
            self.assertEqual(q.body, body)
            self.assertEqual(q.to_dict(), expected)
            self.assertEqual(
                q.line_repr(depth=None),
                (
                    "match_phrase_prefix",
                    'field=message, analyzer="my_analyzer", query="this is a test"',
                ),
            )

    def test_multi_match_clause(self):
        body = {
            "query": "this is a test",
            "fields": ["subject", "message"],
            "type": "best_fields",
        }
        expected = {"multi_match": body}

        q = MultiMatch(
            fields=["subject", "message"], query="this is a test", type="best_fields"
        )
        self.assertEqual(q.body, body)
        self.assertEqual(q.to_dict(), expected)
        self.assertEqual(
            q.line_repr(depth=None), ("multi_match", "fields=['subject', 'message']")
        )

    def test_query_string_clause(self):
        body = {"query": "(new york city) OR (big apple)", "default_field": "content"}
        expected = {"query_string": body}

        q = QueryString(query="(new york city) OR (big apple)", default_field="content")
        self.assertEqual(q.body, body)
        self.assertEqual(q.to_dict(), expected)
        self.assertEqual(
            q.line_repr(depth=None),
            (
                "query_string",
                'default_field="content", query="(new york city) OR (big apple)"',
            ),
        )

    def test_simple_string_clause(self):
        body = {"query": "(new york city) OR (big apple)", "default_field": "content"}
        expected = {"simple_string": body}

        q = SimpleQueryString(
            query="(new york city) OR (big apple)", default_field="content"
        )
        self.assertEqual(q.body, body)
        self.assertEqual(q.to_dict(), expected)
        self.assertEqual(
            q.line_repr(depth=None),
            (
                "simple_string",
                'default_field="content", query="(new york city) OR (big apple)"',
            ),
        )



from pandagg.node.aggs import (
    Terms,
    Filter,
    Filters,
    DateHistogram,
    Nested,
    Range,
    Histogram,
)

from tests import PandaggTestCase


class BucketAggNodesTestCase(PandaggTestCase):
    def test_terms(self):
        es_raw_response = {
            "doc_count_error_upper_bound": 0,
            "sum_other_doc_count": 0,
            "buckets": [
                {"key": "electronic", "doc_count": 6},
                {"key": "rock", "doc_count": 3},
                {"key": "jazz", "doc_count": 2},
            ],
        }
        # test extract_buckets
        buckets_iterator = Terms("name", "field").extract_buckets(es_raw_response)
        self.assertTrue(hasattr(buckets_iterator, "__iter__"))
        buckets = list(buckets_iterator)
        self.assertEqual(
            buckets,
            [
                # key -> bucket
                ("electronic", {"doc_count": 6, "key": "electronic"}),
                ("rock", {"doc_count": 3, "key": "rock"}),
                ("jazz", {"doc_count": 2, "key": "jazz"}),
            ],
        )

        # test extract bucket value
        self.assertEqual(
            Terms("name", "field").extract_bucket_value(
                {"doc_count": 6, "key": "electronic"}
            ),
            6,
        )
        self.assertEqual(
            Terms("name", "field").extract_bucket_value(
                {"doc_count": 3, "key": "rock"}
            ),
            3,
        )
        self.assertEqual(
            Terms("name", "field").extract_bucket_value(
                {"doc_count": 2, "key": "jazz"}
            ),
            2,
        )

        # test query dict
        self.assertEqual(
            Terms(field="field", size=10).to_dict(),
            {"terms": {"field": "field", "size": 10}},
        )

    def test_filter(self):
        es_raw_response = {"doc_count": 12, "sub_aggs": {}}
        # test extract_buckets
        buckets_iterator = Filter({}).extract_buckets(es_raw_response)
        self.assertTrue(hasattr(buckets_iterator, "__iter__"))
        buckets = list(buckets_iterator)
        self.assertEqual(
            buckets,
            [
                # key -> bucket
                (None, {"doc_count": 12, "sub_aggs": {}})
            ],
        )

        # test extract bucket value
        self.assertEqual(Filter({}).extract_bucket_value({"doc_count": 12}), 12)

        # test get_filter
        filter_agg = Filter(filter={"term": {"some_path": 1}})
        self.assertEqual(filter_agg.get_filter(None), {"term": {"some_path": 1}})

        # test query dict
        self.assertEqual(filter_agg.to_dict(), {"filter": {"term": {"some_path": 1}}})

    def test_nested(self):
        es_raw_response = {"doc_count": 12, "sub_aggs": {}}
        # test extract_buckets
        buckets_iterator = Nested(path="some_nested").extract_buckets(es_raw_response)
        self.assertTrue(hasattr(buckets_iterator, "__iter__"))
        buckets = list(buckets_iterator)
        self.assertEqual(
            buckets,
            [
                # key -> bucket
                (None, {"doc_count": 12, "sub_aggs": {}})
            ],
        )

        # test extract bucket value
        self.assertEqual(Nested.extract_bucket_value({"doc_count": 12}), 12)

        # test get_filter
        nested_agg = Nested(path="nested_path")
        self.assertEqual(nested_agg.get_filter(None), None)

        # test query dict
        self.assertEqual(nested_agg.to_dict(), {"nested": {"path": "nested_path"}})

    def test_filters(self):
        es_raw_response = {
            "buckets": {
                "my_first_filter": {"doc_count": 1},
                "my_second_filter": {"doc_count": 2},
            }
        }
        # test extract_buckets
        buckets_iterator = Filters(None).extract_buckets(es_raw_response)
        self.assertTrue(hasattr(buckets_iterator, "__iter__"))
        buckets = list(buckets_iterator)
        self.assertEqual(
            buckets,
            [
                # key -> bucket
                ("my_first_filter", {"doc_count": 1}),
                ("my_second_filter", {"doc_count": 2}),
            ],
        )

        # test extract bucket value
        self.assertEqual(Filters(None).extract_bucket_value({"doc_count": 1}), 1)
        self.assertEqual(Filters(None).extract_bucket_value({"doc_count": 2}), 2)

        # test get_filter
        filters_agg = Filters(
            filters={
                "first_bucket": {"term": {"some_path": 1}},
                "second_bucket": {"term": {"some_path": 2}},
            },
            other_bucket=True,
            other_bucket_key="neither_one_nor_two",
        )
        self.assertEqual(
            filters_agg.get_filter("first_bucket"), {"term": {"some_path": 1}}
        )
        expected_others_filter = {
            "bool": {
                "must_not": {
                    "bool": {
                        "should": [
                            {"term": {"some_path": 1}},
                            {"term": {"some_path": 2}},
                        ]
                    }
                }
            }
        }
        self.assertQueryEqual(filters_agg.get_filter("_other_"), expected_others_filter)
        self.assertQueryEqual(
            filters_agg.get_filter("neither_one_nor_two"), expected_others_filter
        )

        self.assertQueryEqual(
            filters_agg.to_dict(),
            {
                "filters": {
                    "filters": {
                        "first_bucket": {"term": {"some_path": 1}},
                        "second_bucket": {"term": {"some_path": 2}},
                    },
                    "other_bucket": True,
                    "other_bucket_key": "neither_one_nor_two",
                }
            },
        )

    def test_range(self):
        query = {
            "range": {
                "field": "price",
                "ranges": [
                    {"to": 100.0},
                    {"from": 100.0, "to": 200.0},
                    {"from": 200.0},
                ],
            }
        }
        es_raw_response = {
            "buckets": [
                {"key": "*-100.0", "to": 100.0, "doc_count": 2},
                {"key": "100.0-200.0", "from": 100.0, "to": 200.0, "doc_count": 2},
                {"key": "200.0-*", "from": 200.0, "doc_count": 3},
            ]
        }

        range_agg = Range(
            field="price",
            ranges=[{"to": 100.0}, {"from": 100.0, "to": 200.0}, {"from": 200.0}],
        )
        self.assertEqual(range_agg.to_dict(), query)

        buckets_iterator = range_agg.extract_buckets(es_raw_response)
        self.assertTrue(hasattr(buckets_iterator, "__iter__"))
        buckets = list(buckets_iterator)
        self.assertEqual(
            buckets,
            [
                # key -> bucket
                ("*-100.0", {"doc_count": 2, "key": "*-100.0", "to": 100.0}),
                (
                    "100.0-200.0",
                    {"doc_count": 2, "from": 100.0, "key": "100.0-200.0", "to": 200.0},
                ),
                ("200.0-*", {"doc_count": 3, "from": 200.0, "key": "200.0-*"}),
            ],
        )

    def test_range_keyed_response(self):
        query = {
            "range": {
                "field": "price",
                "keyed": True,
                "ranges": [
                    {"to": 100.0},
                    {"from": 100.0, "to": 200.0},
                    {"from": 200.0},
                ],
            }
        }
        es_raw_response = {
            "buckets": {
                "*-100.0": {"to": 100.0, "doc_count": 2},
                "100.0-200.0": {"from": 100.0, "to": 200.0, "doc_count": 2},
                "200.0-*": {"from": 200.0, "doc_count": 3},
            }
        }

        range_agg = Range(
            field="price",
            keyed=True,
            ranges=[{"to": 100.0}, {"from": 100.0, "to": 200.0}, {"from": 200.0}],
        )
        self.assertEqual(range_agg.to_dict(), query)

        buckets_iterator = range_agg.extract_buckets(es_raw_response)
        self.assertTrue(hasattr(buckets_iterator, "__iter__"))
        buckets = list(buckets_iterator)
        self.assertEqual(
            buckets,
            [
                # key -> bucket
                ("*-100.0", {"doc_count": 2, "to": 100.0}),
                ("100.0-200.0", {"doc_count": 2, "from": 100.0, "to": 200.0}),
                ("200.0-*", {"doc_count": 3, "from": 200.0}),
            ],
        )

    def test_histogram(self):
        query = {"histogram": {"field": "price", "interval": 50}}
        es_raw_response = {
            "buckets": [
                {"key": 0.0, "doc_count": 1},
                {"key": 50.0, "doc_count": 1},
                {"key": 100.0, "doc_count": 0},
                {"key": 150.0, "doc_count": 2},
                {"key": 200.0, "doc_count": 3},
            ]
        }

        hist_agg = Histogram(field="price", interval=50)
        self.assertEqual(hist_agg.to_dict(), query)
        self.assertEqual(
            hist_agg.get_filter(100), {"range": {"price": {"gte": 100.0, "lt": 150.0}}}
        )

        buckets_iterator = hist_agg.extract_buckets(es_raw_response)
        self.assertTrue(hasattr(buckets_iterator, "__iter__"))
        buckets = list(buckets_iterator)
        self.assertEqual(
            buckets,
            [
                # key -> bucket
                (0.0, {"doc_count": 1, "key": 0.0}),
                (50.0, {"doc_count": 1, "key": 50.0}),
                (100.0, {"doc_count": 0, "key": 100.0}),
                (150.0, {"doc_count": 2, "key": 150.0}),
                (200.0, {"doc_count": 3, "key": 200.0}),
            ],
        )

    def test_date_histogram_key_as_string(self):
        es_raw_response = {
            "doc_count_error_upper_bound": 0,
            "sum_other_doc_count": 0,
            "buckets": [
                {"key_as_string": "2018-01-01", "key": 1514764800000, "doc_count": 6},
                {"key_as_string": "2018-01-08", "key": 1515369600000, "doc_count": 3},
            ],
        }

        date_hist_agg = DateHistogram(
            name="name", field="field", interval="1w", key_as_string=True
        )
        self.assertEqual(
            date_hist_agg.get_filter("2018-01-01"),
            {"range": {"field": {"gte": "2018-01-01", "lt": "2018-01-01||+1w"}}},
        )

        buckets_iterator = date_hist_agg.extract_buckets(es_raw_response)

        self.assertTrue(hasattr(buckets_iterator, "__iter__"))
        buckets = list(buckets_iterator)
        self.assertEqual(
            buckets,
            [
                # key_as_string -> bucket
                (
                    "2018-01-01",
                    {
                        "doc_count": 6,
                        "key": 1514764800000,
                        "key_as_string": "2018-01-01",
                    },
                ),
                (
                    "2018-01-08",
                    {
                        "doc_count": 3,
                        "key": 1515369600000,
                        "key_as_string": "2018-01-08",
                    },
                ),
            ],
        )

        # not using key as string (regular key)
        buckets_iterator = DateHistogram(
            name="name", field="field", interval="1w", key_as_string=False
        ).extract_buckets(es_raw_response)

        self.assertTrue(hasattr(buckets_iterator, "__iter__"))
        buckets = list(buckets_iterator)
        self.assertEqual(
            buckets,
            [
                # key -> bucket
                (
                    1514764800000,
                    {
                        "doc_count": 6,
                        "key": 1514764800000,
                        "key_as_string": "2018-01-01",
                    },
                ),
                (
                    1515369600000,
                    {
                        "doc_count": 3,
                        "key": 1515369600000,
                        "key_as_string": "2018-01-08",
                    },
                ),
            ],
        )
